# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Receives documents from the oplog worker threads and indexes them
    into the backend.

    This file is a document manager for MongoDB, but the intent
    is that this file can be used as an example to add on different backends.
    To extend this to other systems, simply implement the exact same class and
    replace the method definitions with API calls for the desired backend.
    """

import logging
import pymongo

from bson import SON
from gridfs import GridFS

from mongo_connector import errors, constants
from mongo_connector.util import exception_wrapper
from mongo_connector.doc_managers.doc_manager_base import DocManagerBase

wrap_exceptions = exception_wrapper(
    {
        pymongo.errors.ConnectionFailure: errors.ConnectionFailed,
        pymongo.errors.OperationFailure: errors.OperationFailed,
    }
)

LOG = logging.getLogger(__name__)

__version__ = constants.__version__
"""MongoDB DocManager version information

This is packaged with mongo-connector so it shares the same version.
Downstream DocManager implementations should add their package __version__
string here, for example:

__version__ = '0.1.0'
"""


class DocManager(DocManagerBase):
    """The DocManager class creates a connection to the backend engine and
        adds/removes documents, and in the case of rollback, searches for them.

        The reason for storing id/doc pairs as opposed to doc's is so that
        multiple updates to the same doc reflect the most up to date version as
        opposed to multiple, slightly different versions of a doc.

        We are using MongoDB native fields for _id and ns, but we also store
        them as fields in the document, due to compatibility issues.
        """

    def __init__(self, url, **kwargs):
        """ Verify URL and establish a connection.
        """
        try:
            self.mongo = pymongo.MongoClient(url, **kwargs.get("clientOptions", {}))
        except pymongo.errors.InvalidURI:
            raise errors.ConnectionFailed("Invalid URI for MongoDB")
        except pymongo.errors.ConnectionFailure:
            raise errors.ConnectionFailed("Failed to connect to MongoDB")
        self.chunk_size = kwargs.get("chunk_size", constants.DEFAULT_MAX_BULK)
        self.use_single_meta_collection = kwargs.get(
            "use_single_meta_collection", False
        )
        self.meta_collection_name = kwargs.get(
            "meta_collection_name", constants.DEFAULT_META_COLLECTION_NAME
        )
        self.meta_collection_cap_size = kwargs.get(
            "meta_collection_cap_size", constants.DEFAULT_META_COLLECTION_CAP_SIZE
        )

        # The '_id' field has to be unique, so if we will be writing data from
        # different namespaces into single collection, we use a different field
        # for storing the document id.
        self.id_field = "doc_id" if self.use_single_meta_collection else "_id"
        self.meta_database = self.mongo["__mongo_connector"]

        # Create the meta collection as capped if a single meta collection is
        # preferred
        if self.use_single_meta_collection:
            if self.meta_collection_name not in self.meta_database.collection_names():
                self.meta_database.create_collection(
                    self.meta_collection_name,
                    capped=True,
                    size=self.meta_collection_cap_size,
                )
                meta_collection = self.meta_database[self.meta_collection_name]
                meta_collection.create_index(self.id_field)
                meta_collection.create_index([("ns", 1), ("_ts", 1)])

    def _db_and_collection(self, namespace):
        return namespace.split(".", 1)

    def _get_meta_collection(self, namespace):
        if self.use_single_meta_collection:
            return self.meta_collection_name
        else:
            return namespace

    @wrap_exceptions
    def _meta_collections(self):
        """Provides the meta collections currently being used
        """
        if self.use_single_meta_collection:
            yield self.meta_collection_name
        else:
            for name in self.meta_database.collection_names(
                include_system_collections=False
            ):
                yield name

    def stop(self):
        """Stops any running threads
        """
        LOG.info(
            "Mongo DocManager Stopped: If you will not target this system "
            "again with mongo-connector then you may drop the database "
            "__mongo_connector, which holds metadata for Mongo Connector."
        )

    @wrap_exceptions
    def handle_command(self, doc, namespace, timestamp):
        db, _ = self._db_and_collection(namespace)
        if doc.get("dropDatabase"):
            for new_db in self.command_helper.map_db(db):
                self.mongo.drop_database(new_db)

        if doc.get("renameCollection"):
            a = self.command_helper.map_namespace(doc["renameCollection"])
            b = self.command_helper.map_namespace(doc["to"])
            if a and b:
                self.mongo.admin.command("renameCollection", a, to=b)

        if doc.get("create"):
            new_db, coll = self.command_helper.map_collection(db, doc["create"])
            if new_db:
                self.mongo[new_db].create_collection(coll)

        if doc.get("drop"):
            new_db, coll = self.command_helper.map_collection(db, doc["drop"])
            if new_db:
                self.mongo[new_db].drop_collection(coll)

    @wrap_exceptions
    def update(self, document_id, update_spec, namespace, timestamp):
        """Apply updates given in update_spec to the document whose id
        matches that of doc.

        """
        db, coll = self._db_and_collection(namespace)

        meta_collection_name = self._get_meta_collection(namespace)

        self.meta_database[meta_collection_name].replace_one(
            {self.id_field: document_id, "ns": namespace},
            {self.id_field: document_id, "_ts": timestamp, "ns": namespace},
            upsert=True,
        )

        no_obj_error = "No matching object found"
        update_spec.pop("$v", None)
        updated = self.mongo[db].command(
            SON(
                [
                    ("findAndModify", coll),
                    ("query", {"_id": document_id}),
                    ("update", update_spec),
                    ("new", True),
                ]
            ),
            allowable_errors=[no_obj_error],
        )["value"]
        return updated

    @wrap_exceptions
    def upsert(self, doc, namespace, timestamp):
        """Update or insert a document into Mongo
        """
        database, coll = self._db_and_collection(namespace)

        meta_collection_name = self._get_meta_collection(namespace)

        self.meta_database[meta_collection_name].replace_one(
            {self.id_field: doc["_id"], "ns": namespace},
            {self.id_field: doc["_id"], "_ts": timestamp, "ns": namespace},
            upsert=True,
        )

        self.mongo[database][coll].replace_one({"_id": doc["_id"]}, doc, upsert=True)

    @wrap_exceptions
    def bulk_upsert(self, docs, namespace, timestamp):
        def iterate_chunks():
            dbname, collname = self._db_and_collection(namespace)
            collection = self.mongo[dbname][collname]
            meta_collection_name = self._get_meta_collection(namespace)
            meta_collection = self.meta_database[meta_collection_name]
            more_chunks = True
            while more_chunks:
                bulk = collection.initialize_ordered_bulk_op()
                bulk_meta = meta_collection.initialize_ordered_bulk_op()
                for i in range(self.chunk_size):
                    try:
                        doc = next(docs)
                        selector = {"_id": doc["_id"]}
                        bulk.find(selector).upsert().replace_one(doc)
                        meta_selector = {self.id_field: doc["_id"]}
                        bulk_meta.find(meta_selector).upsert().replace_one(
                            {
                                self.id_field: doc["_id"],
                                "ns": namespace,
                                "_ts": timestamp,
                            }
                        )
                    except StopIteration:
                        more_chunks = False
                        if i > 0:
                            yield bulk, bulk_meta
                        break
                if more_chunks:
                    yield bulk, bulk_meta

        for bulk_op, meta_bulk_op in iterate_chunks():
            try:
                bulk_op.execute()
                meta_bulk_op.execute()
            except pymongo.errors.DuplicateKeyError as e:
                LOG.warn("Continuing after DuplicateKeyError: " + str(e))
            except pymongo.errors.BulkWriteError as bwe:
                LOG.error(bwe.details)
                raise

    @wrap_exceptions
    def remove(self, document_id, namespace, timestamp):
        """Removes document from Mongo

        The input is a python dictionary that represents a mongo document.
        The documents has ns and _ts fields.
        """
        database, coll = self._db_and_collection(namespace)

        meta_collection = self._get_meta_collection(namespace)

        doc2 = self.meta_database[meta_collection].find_one_and_delete(
            {self.id_field: document_id}
        )
        if doc2 and doc2.get("gridfs_id"):
            GridFS(self.mongo[database], coll).delete(doc2["gridfs_id"])
        else:
            self.mongo[database][coll].delete_one({"_id": document_id})

    @wrap_exceptions
    def insert_file(self, f, namespace, timestamp):
        database, coll = self._db_and_collection(namespace)

        id = GridFS(self.mongo[database], coll).put(f, filename=f.filename)

        meta_collection = self._get_meta_collection(namespace)

        self.meta_database[meta_collection].replace_one(
            {self.id_field: f._id, "ns": namespace},
            {self.id_field: f._id, "_ts": timestamp, "ns": namespace, "gridfs_id": id},
            upsert=True,
        )

    @wrap_exceptions
    def search(self, start_ts, end_ts):
        """Called to query Mongo for documents in a time range.
        """
        for meta_collection_name in self._meta_collections():
            meta_coll = self.meta_database[meta_collection_name]
            for ts_ns_doc in meta_coll.find(
                {"_ts": {"$lte": end_ts, "$gte": start_ts}}
            ):
                yield ts_ns_doc

    def commit(self):
        """ Performs a commit
        """
        return

    @wrap_exceptions
    def get_last_doc(self):
        """Returns the last document stored in Mongo.
        """

        def docs_by_ts():
            for meta_collection_name in self._meta_collections():
                meta_coll = self.meta_database[meta_collection_name]
                for ts_ns_doc in meta_coll.find(limit=-1).sort("_ts", -1):
                    yield ts_ns_doc

        return max(docs_by_ts(), key=lambda x: x["_ts"])



"""
Install and Uninstall Mongo Connector as a Linux system daemon
"""

import platform
import os
import shutil
import pathlib

import autocommand
import importlib_resources as res


def check_env():
    if platform.system() != "Linux":
        print("Must be running Linux")
        raise SystemExit(1)
    if os.geteuid() > 0:
        print("Must be root user")
        raise SystemExit(2)


log_path = pathlib.Path("/var/log/mongo-connector")
init_script = pathlib.Path("/etc/init.d/mongo-connector")
config_file = pathlib.Path("/etc/mongo-connector.json")
package = "mongo_connector.service"


def install():
    log_path.mkdir(exist_ok=True)
    init_script.dirname().makedirs_p()
    with res.path(package, "config.json") as config_src_path:
        shutil.copyfile(config_src_path, config_file)
    with res.path(package, "System V init") as init_src_path:
        shutil.copyfile(init_src_path, init_script)


def uninstall():
    shutil.rmtree(log_path)
    config_file.unlink()
    init_script.unlink()


@autocommand.autocommand(__name__)
def run(command):
    check_env()
    globals()[command]()



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test replication of commands
"""

import sys

sys.path[0:0] = [""]  # noqa

import pymongo

from mongo_connector.namespace_config import NamespaceConfig
from mongo_connector.command_helper import CommandHelper
from mongo_connector.doc_managers.doc_manager_base import DocManagerBase
from mongo_connector.locking_dict import LockingDict
from mongo_connector.oplog_manager import OplogThread
from mongo_connector.test_utils import assert_soon, close_client, ReplicaSetSingle
from tests import unittest


class CommandLoggerDocManager(DocManagerBase):
    def __init__(self, url=None, **kwargs):
        self.commands = []
        self.command_helper = None

    def stop(self):
        pass

    def upsert(self, doc, namespace, timestamp):
        pass

    def remove(self, document_id, namespace, timestamp):
        pass

    def commit(self):
        pass

    def handle_command(self, doc, namespace, timestamp):
        db, _ = namespace.split(".", 1)
        if doc.get("dropDatabase"):
            if self.command_helper.map_db(db):
                self.commands.append(doc)

        if doc.get("renameCollection"):
            a = self.command_helper.map_namespace(doc["renameCollection"])
            b = self.command_helper.map_namespace(doc["to"])
            if a and b:
                self.commands.append(doc)

        if doc.get("create"):
            new_db, coll = self.command_helper.map_collection(db, doc["create"])
            if new_db:
                self.commands.append(doc)

        if doc.get("drop"):
            new_db, coll = self.command_helper.map_collection(db, doc["drop"])
            if new_db:
                self.commands.append(doc)


class TestCommandReplication(unittest.TestCase):
    def setUp(self):
        self.repl_set = ReplicaSetSingle().start()
        self.primary_conn = self.repl_set.client()
        self.oplog_progress = LockingDict()
        self.opman = None

    def tearDown(self):
        try:
            if self.opman:
                self.opman.join()
        except RuntimeError:
            pass
        close_client(self.primary_conn)
        self.repl_set.stop()

    def initOplogThread(self, namespace_set=None):
        self.docman = CommandLoggerDocManager()
        namespace_config = NamespaceConfig(namespace_set=namespace_set)

        self.docman.command_helper = CommandHelper(namespace_config)
        self.opman = OplogThread(
            primary_client=self.primary_conn,
            doc_managers=(self.docman,),
            oplog_progress_dict=self.oplog_progress,
            namespace_config=namespace_config,
            collection_dump=False,
        )
        self.opman.start()

    def test_command_helper(self):

        mapping = {"a.x": "b.x", "a.y": "c.y"}

        helper = CommandHelper(
            NamespaceConfig(
                namespace_set=list(mapping) + ["a.z"], namespace_options=mapping
            )
        )

        self.assertEqual(set(helper.map_db("a")), set(["a", "b", "c"]))
        self.assertEqual(helper.map_db("d"), [])

        self.assertEqual(helper.map_namespace("a.x"), "b.x")
        self.assertEqual(helper.map_namespace("a.z"), "a.z")
        self.assertEqual(helper.map_namespace("d.x"), None)

        self.assertEqual(helper.map_collection("a", "x"), ("b", "x"))
        self.assertEqual(helper.map_collection("a", "z"), ("a", "z"))
        self.assertEqual(helper.map_collection("d", "x"), (None, None))

    def test_create_collection(self):
        self.initOplogThread()
        pymongo.collection.Collection(self.primary_conn["test"], "test", create=True)
        assert_soon(lambda: self.docman.commands)
        command = self.docman.commands[0]
        self.assertEqual(command["create"], "test")

    def test_create_collection_skipped(self):
        self.initOplogThread(["test.test"])

        pymongo.collection.Collection(self.primary_conn["test2"], "test2", create=True)
        pymongo.collection.Collection(self.primary_conn["test"], "test", create=True)

        assert_soon(lambda: self.docman.commands)
        self.assertEqual(len(self.docman.commands), 1)
        command = self.docman.commands[0]
        self.assertEqual(command["create"], "test")

    def test_drop_collection(self):
        self.initOplogThread()
        coll = pymongo.collection.Collection(
            self.primary_conn["test"], "test", create=True
        )
        coll.drop()
        assert_soon(lambda: len(self.docman.commands) == 2)
        self.assertEqual(self.docman.commands[1], {"drop": "test"})

    def test_drop_database(self):
        self.initOplogThread()
        pymongo.collection.Collection(self.primary_conn["test"], "test", create=True)
        self.primary_conn.drop_database("test")
        assert_soon(lambda: len(self.docman.commands) == 2)
        self.assertEqual(self.docman.commands[1], {"dropDatabase": 1})

    def test_rename_collection(self):
        self.initOplogThread()
        coll = pymongo.collection.Collection(
            self.primary_conn["test"], "test", create=True
        )
        coll.rename("test2")
        assert_soon(lambda: len(self.docman.commands) == 2)
        self.assertEqual(self.docman.commands[1].get("renameCollection"), "test.test")
        self.assertEqual(self.docman.commands[1].get("to"), "test.test2")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python
import pymongo
import sys
from mongo_connector import util

mongo_url = "mongodb://localhost:27017"
if len(sys.argv) == 1:
    print(
        "First argument is mongodb connection string, i.e. "
        "localhost:27017. Assuming localhost:27017..."
    )
if len(sys.argv) >= 2:
    mongo_url = sys.argv[1]


client = pymongo.MongoClient(mongo_url)
rs_name = client.admin.command("ismaster")["setName"]
print("Found Replica Set name: {}".format(str(rs_name)))

print("Now checking for the latest oplog entry...")
oplog = client.local.oplog.rs
last_oplog = oplog.find().sort("$natural", pymongo.DESCENDING).limit(-1).next()
print("Found the last oplog ts: {}".format(str(last_oplog["ts"])))
last_ts = util.bson_ts_to_long(last_oplog["ts"])
out_str = '["{}", {}]'.format(str(rs_name), str(last_ts))

print(
    "Writing all to file oplog.timestamp.last in the format "
    "required for mongo-connector"
)
f = open("./oplog.timestamp.last", "w")
f.write(out_str)
f.close()
print("All done!")



"""Deprecated compat library, retained for doc managers"""

import sys
import warnings

from urllib.request import Request
from urllib.request import urlopen
from urllib.parse import urlencode
from urllib.error import URLError
from urllib.error import HTTPError

__all__ = [
    'Request', 'urlopen', 'urlencode', 'URLError', 'HTTPError',
    'reraise', 'is_string', 'u', 'PY3',
]

warnings.warn(
    "mongo_connector.compat should not be used",
    DeprecationWarning,
)

PY3 = sys.version_info[0] == 3


def reraise(exctype, value, trace=None):
    raise exctype(str(value)).with_traceback(trace)


def is_string(x):
    return isinstance(x, str)


u = str



# Copyright 2013-2016 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import time

from pymongo import MongoClient, ReadPreference

sys.path[0:0] = [""]  # noqa

from mongo_connector.connector import Connector, get_mininum_mongodb_version
from mongo_connector.doc_managers.doc_manager_simulator import DocManager
from mongo_connector.test_utils import (
    assert_soon,
    db_user,
    db_password,
    ShardedClusterSingle,
)
from mongo_connector.version import Version

from tests import unittest, SkipTest


class ShardedConnectorTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if db_user and db_password:
            auth_args = dict(auth_username=db_user, auth_key=db_password)
        else:
            auth_args = {}
        cls.cluster = ShardedClusterSingle().start()
        cls.main_uri = cls.cluster.uri + "/?readPreference=primaryPreferred"
        cls.dm = DocManager()
        cls.connector = Connector(
            mongo_address=cls.main_uri, doc_managers=[cls.dm], **auth_args
        )
        cls.connector.start()
        assert_soon(
            lambda: len(cls.connector.shard_set) == 2,
            message="connector failed to find both shards!",
        )

    @classmethod
    def tearDownClass(cls):
        cls.connector.join()
        try:
            os.unlink("oplog.timestamp")
        except OSError:
            pass
        cls.cluster.stop()


class TestConnectorSharded(ShardedConnectorTestCase):
    def test_connector(self):
        """Test whether the connector initiates properly
        """
        # Make sure get_mininum_mongodb_version returns the current version.
        self.assertEqual(
            Version.from_client(self.cluster.client()), get_mininum_mongodb_version()
        )

    def test_mongos_connection_failure(self):
        """Test that the connector handles temporary mongos failure"""
        client = self.cluster.client()
        coll = client.test.test
        coll.insert_one({"doc": 1})
        assert_soon(lambda: len(self.dm._search()) == 1)

        # Temporarily shutdown the connector mongos connection
        self.connector.main_conn.close()
        time.sleep(5)
        coll.insert_one({"doc": 2})
        assert_soon(lambda: len(self.dm._search()) == 2)

        # Bring mongos back online
        self.connector.main_conn = self.connector.create_authed_client()
        coll.insert_one({"doc": 3})
        assert_soon(lambda: len(self.dm._search()) == 3)
        client.close()

    def test_connector_uri_options(self):
        """Ensure the connector passes URI options to newly created clients.
        """
        expected_client = MongoClient(self.main_uri)

        def assert_client_options(client):
            self.assertEqual(client.read_preference, expected_client.read_preference)
            self.assertEqual(client.read_preference, ReadPreference.PRIMARY_PREFERRED)

        assert_client_options(self.connector.create_authed_client())
        assert_client_options(self.connector.main_conn)
        for oplog_thread in self.connector.shard_set.values():
            assert_client_options(oplog_thread.primary_client)


class TestConnectorShardedAuth(ShardedConnectorTestCase):
    @classmethod
    def setUp(cls):
        if not (db_user and db_password):
            raise SkipTest("Need to set a user/password to test this.")
        super(TestConnectorShardedAuth, cls).setUpClass()

    def test_start_with_auth(self):
        # Insert some documents into the sharded cluster.  These
        # should go to the DocManager, and the connector should not
        # have an auth failure.
        self.cluster.client().test.test.insert_one({"auth_failure": False})
        assert_soon(lambda: len(self.dm._search()) > 0)


if __name__ == "__main__":
    unittest.main()



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import optparse
import sys

from mongo_connector import errors
from mongo_connector.constants import __version__


def default_apply_function(option, cli_values):
    first_value = list(cli_values.values())[0]
    if first_value is not None:
        option.value = first_value


class Option(object):
    """A config file option which can be overwritten on the command line.

    config_key is the corresponding field in the JSON config file.

    apply_function has the following signature:
    def apply_function(option, cli_values):
        # modify option.value ...

    When apply_function is invoked, option.value will be set to the
    value given in the config file (or the default value).

    apply_function reads the cli_values and modifies option.value accordingly
    """

    def __init__(
        self,
        config_key=None,
        default=None,
        type=None,
        apply_function=default_apply_function,
    ):
        self.config_key = config_key
        self.value = default
        self.type = type
        self.apply_function = apply_function

        self.cli_names = []
        self.cli_options = []

    def validate_type(self):
        if self.type == str:
            return isinstance(self.value, str)
        else:
            return isinstance(self.value, self.type)

    def add_cli(self, *args, **kwargs):
        """Add a command line argument.

        All of the given arguments will be passed directly to
        optparse.OptionParser().add_option
        """
        self.cli_options.append((args, kwargs))


class Config(object):
    """Manages command line application configuration.

    conf = Config(options)
    conf.parse_args()
    value = conf['key']
    value2 = conf['key1.key2'] # same as conf['key1']['key2']
    """

    def __init__(self, options):
        self.options = options

        self.config_key_to_option = dict(
            [(option.config_key, option) for option in self.options]
        )

    def parse_args(self, argv=None):
        """Parses command line arguments from stdin (or given argv).

        Does the following:
        1. Parses command line arguments
        2. Loads config file into options (if config file specified)
        3. calls option.apply_function with the parsed cli_values
        """

        # parse the command line options
        parser = optparse.OptionParser(version="%prog version: " + __version__)
        for option in self.options:
            for args, kwargs in option.cli_options:
                cli_option = parser.add_option(*args, **kwargs)
                option.cli_names.append(cli_option.dest)
        parsed_options, args = parser.parse_args(argv)
        if args:
            raise errors.InvalidConfiguration(
                "The following command line arguments are not recognized: "
                + ", ".join(args)
            )

        # load the config file
        if parsed_options.config_file:
            try:
                with open(parsed_options.config_file) as f:
                    self.load_json(f.read())
            except (OSError, IOError, ValueError) as exc:
                tb = sys.exc_info()[2]
                raise errors.InvalidConfiguration(str(exc)).with_traceback(tb)

        # apply the command line arguments
        values = parsed_options.__dict__
        for option in self.options:
            option.apply_function(
                option, dict((k, values.get(k)) for k in option.cli_names)
            )

    def __getitem__(self, key):
        keys = key.split(".")
        cur = self.config_key_to_option[keys[0]].value
        for k in keys[1:]:
            if cur is not None:
                if isinstance(cur, dict):
                    cur = cur.get(k)
                else:
                    cur = None
        return cur

    def load_json(self, text):
        parsed_config = json.loads(text)
        for k in parsed_config:
            option = self.config_key_to_option.get(k)
            if option:
                # load into option.value
                if isinstance(parsed_config[k], dict):
                    for k2 in parsed_config[k]:
                        option.value[k2] = parsed_config[k][k2]
                else:
                    option.value = parsed_config[k]

                # type check
                if not option.validate_type():
                    raise errors.InvalidConfiguration(
                        "%s should be a %r, %r was given!"
                        % (
                            option.config_key,
                            option.type.__name__,
                            type(option.value).__name__,
                        )
                    )
            else:
                if not k.startswith("__"):
                    logging.warning("Unrecognized option: %s" % k)



# Copyright 2013-2016 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test Mongo Connector's behavior when its source MongoDB system is
experiencing a rollback.

"""

import os
import sys
import time

from pymongo.read_preferences import ReadPreference
from pymongo.write_concern import WriteConcern

sys.path[0:0] = [""]  # noqa

from mongo_connector.doc_managers.doc_manager_simulator import DocManager
from mongo_connector.locking_dict import LockingDict
from mongo_connector.namespace_config import NamespaceConfig
from mongo_connector.oplog_manager import OplogThread
from mongo_connector.test_utils import (
    assert_soon,
    close_client,
    ReplicaSet,
    STRESS_COUNT,
)
from mongo_connector.util import retry_until_ok
from tests import unittest


class TestRollbacks(unittest.TestCase):
    def tearDown(self):
        close_client(self.primary_conn)
        close_client(self.secondary_conn)
        try:
            self.opman.join()
        except RuntimeError:
            # OplogThread may not have been started
            pass
        self.repl_set.stop()

    def setUp(self):
        # Create a new oplog progress file
        try:
            os.unlink("oplog.timestamp")
        except OSError:
            pass
        open("oplog.timestamp", "w").close()

        # Start a replica set
        self.repl_set = ReplicaSet().start()
        # Connection to the replica set as a whole
        self.main_conn = self.repl_set.client()
        # Connection to the primary specifically
        self.primary_conn = self.repl_set.primary.client()
        # Connection to the secondary specifically
        self.secondary_conn = self.repl_set.secondary.client(
            read_preference=ReadPreference.SECONDARY_PREFERRED
        )

        # Wipe any test data
        self.main_conn.drop_database("test")

        # Oplog thread
        doc_manager = DocManager()
        oplog_progress = LockingDict()
        self.opman = OplogThread(
            primary_client=self.main_conn,
            doc_managers=(doc_manager,),
            oplog_progress_dict=oplog_progress,
            namespace_config=NamespaceConfig(namespace_set=["test.mc"]),
        )

    def test_single_target(self):
        """Test with a single replication target"""

        self.opman.start()

        # Insert first document with primary up
        self.main_conn["test"]["mc"].insert_one({"i": 0})
        self.assertEqual(self.primary_conn["test"]["mc"].find().count(), 1)

        # Make sure the insert is replicated
        secondary = self.secondary_conn
        assert_soon(
            lambda: secondary["test"]["mc"].count() == 1,
            "first write didn't replicate to secondary",
        )

        # Kill the primary
        self.repl_set.primary.stop(destroy=False)

        # Wait for the secondary to be promoted
        assert_soon(lambda: secondary["admin"].command("isMaster")["ismaster"])

        # Insert another document. This will be rolled back later
        retry_until_ok(self.main_conn["test"]["mc"].insert_one, {"i": 1})
        self.assertEqual(secondary["test"]["mc"].count(), 2)

        # Wait for replication to doc manager
        assert_soon(
            lambda: len(self.opman.doc_managers[0]._search()) == 2,
            "not all writes were replicated to doc manager",
        )

        # Kill the new primary
        self.repl_set.secondary.stop(destroy=False)

        # Start both servers back up
        self.repl_set.primary.start()
        primary_admin = self.primary_conn["admin"]
        assert_soon(
            lambda: primary_admin.command("isMaster")["ismaster"],
            "restarted primary never resumed primary status",
        )
        self.repl_set.secondary.start()
        assert_soon(
            lambda: retry_until_ok(secondary.admin.command, "replSetGetStatus")[
                "myState"
            ]
            == 2,
            "restarted secondary never resumed secondary status",
        )
        assert_soon(
            lambda: retry_until_ok(self.main_conn.test.mc.find().count) > 0,
            "documents not found after primary/secondary restarted",
        )

        # Only first document should exist in MongoDB
        self.assertEqual(self.main_conn["test"]["mc"].count(), 1)
        self.assertEqual(self.main_conn["test"]["mc"].find_one()["i"], 0)

        # Same case should hold for the doc manager
        doc_manager = self.opman.doc_managers[0]
        assert_soon(
            lambda: len(doc_manager._search()) == 1,
            "documents never rolled back in doc manager.",
        )
        self.assertEqual(doc_manager._search()[0]["i"], 0)

        # cleanup
        self.opman.join()

    def test_many_targets(self):
        """Test with several replication targets"""

        # OplogThread has multiple doc managers
        doc_managers = [DocManager(), DocManager(), DocManager()]
        self.opman.doc_managers = doc_managers

        self.opman.start()

        # Insert a document into each namespace
        self.main_conn["test"]["mc"].insert_one({"i": 0})
        self.assertEqual(self.primary_conn["test"]["mc"].count(), 1)

        # Make sure the insert is replicated
        secondary = self.secondary_conn
        assert_soon(
            lambda: secondary["test"]["mc"].count() == 1,
            "first write didn't replicate to secondary",
        )

        # Kill the primary
        self.repl_set.primary.stop(destroy=False)

        # Wait for the secondary to be promoted
        assert_soon(
            lambda: secondary.admin.command("isMaster")["ismaster"],
            "secondary was never promoted",
        )

        # Insert more documents. This will be rolled back later
        # Some of these documents will be manually removed from
        # certain doc managers, to emulate the effect of certain
        # target systems being ahead/behind others
        secondary_ids = []
        for i in range(1, 10):
            secondary_ids.append(
                retry_until_ok(
                    self.main_conn["test"]["mc"].insert_one, {"i": i}
                ).inserted_id
            )
        self.assertEqual(self.secondary_conn["test"]["mc"].count(), 10)

        # Wait for replication to the doc managers
        def docmans_done():
            for dm in self.opman.doc_managers:
                if len(dm._search()) != 10:
                    return False
            return True

        assert_soon(docmans_done, "not all writes were replicated to doc managers")

        # Remove some documents from the doc managers to simulate
        # uneven replication
        ts = self.opman.doc_managers[0].get_last_doc()["_ts"]
        for id in secondary_ids[8:]:
            self.opman.doc_managers[1].remove(id, "test.mc", ts)
        for id in secondary_ids[2:]:
            self.opman.doc_managers[2].remove(id, "test.mc", ts)

        # Kill the new primary
        self.repl_set.secondary.stop(destroy=False)

        # Start both servers back up
        self.repl_set.primary.start()
        primary_admin = self.primary_conn["admin"]
        assert_soon(
            lambda: primary_admin.command("isMaster")["ismaster"],
            "restarted primary never resumed primary status",
        )
        self.repl_set.secondary.start()
        assert_soon(
            lambda: retry_until_ok(secondary.admin.command, "replSetGetStatus")[
                "myState"
            ]
            == 2,
            "restarted secondary never resumed secondary status",
        )
        assert_soon(
            lambda: retry_until_ok(self.primary_conn.test.mc.find().count) > 0,
            "documents not found after primary/secondary restarted",
        )

        # Only first document should exist in MongoDB
        self.assertEqual(self.primary_conn["test"]["mc"].count(), 1)
        self.assertEqual(self.primary_conn["test"]["mc"].find_one()["i"], 0)

        # Give OplogThread some time to catch up
        time.sleep(10)

        # Same case should hold for the doc managers
        for dm in self.opman.doc_managers:
            self.assertEqual(len(dm._search()), 1)
            self.assertEqual(dm._search()[0]["i"], 0)

        self.opman.join()

    def test_deletions(self):
        """Test rolling back 'd' operations"""

        self.opman.start()

        # Insert a document, wait till it replicates to secondary
        self.main_conn["test"]["mc"].insert_one({"i": 0})
        self.main_conn["test"]["mc"].insert_one({"i": 1})
        self.assertEqual(self.primary_conn["test"]["mc"].find().count(), 2)
        assert_soon(
            lambda: self.secondary_conn["test"]["mc"].count() == 2,
            "first write didn't replicate to secondary",
        )

        # Kill the primary, wait for secondary to be promoted
        self.repl_set.primary.stop(destroy=False)
        assert_soon(
            lambda: self.secondary_conn["admin"].command("isMaster")["ismaster"]
        )

        # Delete first document
        retry_until_ok(self.main_conn["test"]["mc"].delete_one, {"i": 0})
        self.assertEqual(self.secondary_conn["test"]["mc"].count(), 1)

        # Wait for replication to doc manager
        assert_soon(
            lambda: len(self.opman.doc_managers[0]._search()) == 1,
            "delete was not replicated to doc manager",
        )

        # Kill the new primary
        self.repl_set.secondary.stop(destroy=False)

        # Start both servers back up
        self.repl_set.primary.start()
        primary_admin = self.primary_conn["admin"]
        assert_soon(
            lambda: primary_admin.command("isMaster")["ismaster"],
            "restarted primary never resumed primary status",
        )
        self.repl_set.secondary.start()
        assert_soon(
            lambda: retry_until_ok(
                self.secondary_conn.admin.command, "replSetGetStatus"
            )["myState"]
            == 2,
            "restarted secondary never resumed secondary status",
        )

        # Both documents should exist in mongo
        assert_soon(lambda: retry_until_ok(self.main_conn["test"]["mc"].count) == 2)

        # Both document should exist in doc manager
        doc_manager = self.opman.doc_managers[0]
        assert_soon(
            lambda: len(list(doc_manager._search())) == 2,
            ("Expected two documents, but got: %r" % list(doc_manager._search())),
        )

        self.opman.join()

    def test_stressed_rollback(self):
        """Stress test for a rollback with many documents."""
        self.opman.start()

        c = self.main_conn.test.mc
        docman = self.opman.doc_managers[0]
        c2 = c.with_options(write_concern=WriteConcern(w=2))
        c2.insert_many([{"i": i} for i in range(STRESS_COUNT)])
        assert_soon(lambda: c2.count() == STRESS_COUNT)

        def condition():
            return len(docman._search()) == STRESS_COUNT

        assert_soon(
            condition,
            (
                "Was expecting %d documents in DocManager, "
                "but %d found instead." % (STRESS_COUNT, len(docman._search()))
            ),
        )

        primary_conn = self.repl_set.primary.client()
        self.repl_set.primary.stop(destroy=False)
        new_primary_conn = self.repl_set.secondary.client()

        admin = new_primary_conn.admin
        assert_soon(lambda: retry_until_ok(admin.command, "isMaster")["ismaster"])

        retry_until_ok(
            c.insert_many, [{"i": str(STRESS_COUNT + i)} for i in range(STRESS_COUNT)]
        )

        self.repl_set.secondary.stop(destroy=False)

        self.repl_set.primary.start()
        admin = primary_conn.admin
        assert_soon(lambda: retry_until_ok(admin.command, "isMaster")["ismaster"])
        self.repl_set.secondary.start()

        assert_soon(lambda: retry_until_ok(c.count) == STRESS_COUNT)
        assert_soon(
            condition,
            (
                "Was expecting %d documents in DocManager, "
                "but %d found instead." % (STRESS_COUNT, len(docman._search()))
            ),
        )

        self.opman.join()


if __name__ == "__main__":
    unittest.main()



# Copyright 2016 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test oplog manager methods with wildcard namespaces
"""

import itertools
import re
import sys
import time

import bson
import pymongo

sys.path[0:0] = [""]  # noqa

from mongo_connector.doc_managers.doc_manager_simulator import DocManager
from mongo_connector.locking_dict import LockingDict
from mongo_connector.namespace_config import NamespaceConfig
from mongo_connector.oplog_manager import OplogThread
from mongo_connector.test_utils import assert_soon, close_client, ReplicaSetSingle
from tests import unittest


class TestOplogManager(unittest.TestCase):
    """Defines all the testing methods, as well as a method that sets up the
    cluster
    """

    def setUp(self):
        self.repl_set = ReplicaSetSingle().start()
        self.primary_conn = self.repl_set.client()
        self.oplog_coll = self.primary_conn.local["oplog.rs"]

    def reset_opman(self, include_ns=None, exclude_ns=None, dest_mapping=None):
        self.namespace_config = NamespaceConfig(
            namespace_set=include_ns,
            ex_namespace_set=exclude_ns,
            namespace_options=dest_mapping,
        )
        self.opman = OplogThread(
            primary_client=self.primary_conn,
            doc_managers=(DocManager(),),
            oplog_progress_dict=LockingDict(),
            namespace_config=self.namespace_config,
        )

    def init_dbs(self):
        # includedb1.* & includedb2.includecol1 are interested collections
        self.primary_conn["includedb1"]["includecol1"].insert_many(
            [{"idb1col1": i} for i in range(1, 3)]
        )

        self.primary_conn["includedb1"]["includecol2"].insert_many(
            [{"idb1col2": i} for i in range(1, 3)]
        )

        self.primary_conn["includedb2"]["includecol1"].insert_many(
            [{"idb2col1": i} for i in range(1, 3)]
        )

        # the others are not interested collections
        self.primary_conn["includedb2"]["excludecol2"].insert_many(
            [{"idb2col2": i} for i in range(1, 3)]
        )

        self.primary_conn["excludedb3"]["excludecol1"].insert_many(
            [{"idb3col1": i} for i in range(1, 3)]
        )

    def tearDown(self):
        try:
            self.opman.join()
        except RuntimeError:
            pass  # OplogThread may not have been started

        for db in self.primary_conn.database_names():
            if db != "local":
                self.primary_conn.drop_database(db)
        close_client(self.primary_conn)
        self.repl_set.stop()

    def test_get_oplog_cursor(self):
        """Test the get_oplog_cursor method"""

        # Put something in the dbs
        self.init_dbs()

        # timestamp is None - all oplog entries excluding no-ops are returned.
        # wildcard include case no impact the result
        self.reset_opman(["includedb1.*", "includedb2.includecol1"], [], {})

        got_cursor = self.opman.get_oplog_cursor(None)
        oplog_cursor = self.oplog_coll.find({"op": {"$ne": "n"}})
        self.assertNotEqual(got_cursor, None)
        self.assertEqual(got_cursor.count(), oplog_cursor.count())

        # wildcard exclude case no impact the result
        self.reset_opman([], ["includedb2.excludecol2", "excludedb3.*"], {})

        got_cursor = self.opman.get_oplog_cursor(None)
        oplog_cursor = self.oplog_coll.find({"op": {"$ne": "n"}})
        self.assertNotEqual(got_cursor, None)
        self.assertEqual(got_cursor.count(), oplog_cursor.count())

        # earliest entry is the only one at/after timestamp
        doc = {"ts": bson.Timestamp(1000, 0), "idb1col1": 1}
        self.primary_conn["includedb1"]["includecol1"].insert_one(doc)
        latest_timestamp = self.opman.get_last_oplog_timestamp()
        cursor = self.opman.get_oplog_cursor(latest_timestamp)
        self.assertNotEqual(cursor, None)
        self.assertEqual(cursor.count(), 1)
        next_entry_id = next(cursor)["o"]["_id"]
        retrieved = self.primary_conn.includedb1.includecol1.find_one(next_entry_id)
        self.assertEqual(retrieved, doc)

        # many entries before and after timestamp
        self.primary_conn["includedb1"]["includecol1"].insert_many(
            [{"idb1col1": i} for i in range(2, 1002)]
        )
        oplog_cursor = self.oplog_coll.find(
            {"op": {"$ne": "n"}, "ns": {"$not": re.compile(r"\.(system|\$cmd)")}},
            sort=[("ts", pymongo.ASCENDING)],
        )

        # initial insert + 1000 more inserts
        self.assertEqual(oplog_cursor.count(), 11 + 1000)
        pivot = oplog_cursor.skip(400).limit(-1)[0]

        goc_cursor = self.opman.get_oplog_cursor(pivot["ts"])
        self.assertEqual(goc_cursor.count(), 11 + 1000 - 400)

    def test_get_last_oplog_timestamp(self):
        """Test the get_last_oplog_timestamp method"""

        # empty oplog case has been tested in test_oplog_manager.py,
        # skip that here.

        # Put something in the dbs
        self.init_dbs()

        # Test non-empty oplog
        self.reset_opman(["includedb1.*", "includedb2.includecol1"], [], {})
        self.opman.oplog = self.primary_conn["local"]["oplog.rs"]
        for i in range(1000):
            self.primary_conn["includedb1"]["includecol1"].insert_one(
                {"idb1col1": i + 500}
            )
        oplog = self.primary_conn["local"]["oplog.rs"]
        oplog = (
            oplog.find({"op": {"$ne": "n"}})
            .sort("$natural", pymongo.DESCENDING)
            .limit(-1)[0]
        )
        self.assertEqual(self.opman.get_last_oplog_timestamp(), oplog["ts"])

    def test_dump_collection(self):
        """Test the dump_collection method

        Cases:

        1. no namespace set is set
        2. include namespace set is set
        3. exclude namespace set is set

        empty oplog case has been tested in test_oplog_manager.py,
        skip that here.
        """

        # Put something in the dbs
        self.init_dbs()

        # no namespace set is set
        self.reset_opman([], [], {})
        self.opman.oplog = self.primary_conn["local"]["oplog.rs"]
        last_ts = self.opman.get_last_oplog_timestamp()
        self.assertEqual(last_ts, self.opman.dump_collection())
        self.assertEqual(len(self.opman.doc_managers[0]._search()), 10)

        # include namespace set is set
        self.reset_opman(["includedb1.*", "includedb2.includecol1"], [], {})
        self.opman.oplog = self.primary_conn["local"]["oplog.rs"]
        last_ts = self.opman.get_last_oplog_timestamp()
        self.assertEqual(last_ts, self.opman.dump_collection())
        self.assertEqual(len(self.opman.doc_managers[0]._search()), 6)

        # exclude namespace set is set
        self.reset_opman([], ["includedb2.excludecol2", "excludedb3.*"], {})
        self.opman.oplog = self.primary_conn["local"]["oplog.rs"]
        last_ts = self.opman.get_last_oplog_timestamp()
        self.assertEqual(last_ts, self.opman.dump_collection())
        self.assertEqual(len(self.opman.doc_managers[0]._search()), 6)

    def test_dump_collection_with_error(self):
        """Test the dump_collection method with invalid documents.

        Cases:

        1. non-empty oplog, continue_on_error=True, invalid documents
        """

        self.reset_opman(["includedb1.*", "includedb2.includecol1"], [], {})
        # non-empty oplog, continue_on_error=True, invalid documents
        self.opman.continue_on_error = True
        self.opman.oplog = self.primary_conn["local"]["oplog.rs"]

        docs = [{"a": i} for i in range(100)]
        for i in range(50, 60):
            docs[i]["_upsert_exception"] = True
        self.primary_conn["includedb1"]["includecol3"].insert_many(docs)

        last_ts = self.opman.get_last_oplog_timestamp()
        self.assertEqual(last_ts, self.opman.dump_collection())
        docs = self.opman.doc_managers[0]._search()
        docs = list(filter(lambda doc: "a" in doc, docs))
        docs.sort(key=lambda doc: doc["a"])

        self.assertEqual(len(docs), 90)
        expected_a = itertools.chain(range(0, 50), range(60, 100))
        for doc, correct_a in zip(docs, expected_a):
            self.assertEqual(doc["a"], correct_a)

    def test_init_cursor(self):
        """Test the init_cursor method

        Cases:

        1. no last checkpoint, no collection dump
        2. no last checkpoint, collection dump ok and stuff to dump
        3. no last checkpoint, nothing to dump, stuff in oplog
        4. no last checkpoint, nothing to dump, nothing in oplog
        5. no last checkpoint, no collection dump, stuff in oplog
        6. last checkpoint exists
        7. last checkpoint is behind
        """

        # N.B. these sub-cases build off of each other and cannot be re-ordered
        # without side-effects

        self.reset_opman(["includedb1.*", "includedb2.includecol1"], [], {})

        # No last checkpoint, no collection dump, nothing in oplog
        # "change oplog collection" to put nothing in oplog
        self.opman.oplog = self.primary_conn["includedb1"]["emptycollection"]
        self.opman.collection_dump = False
        self.assertTrue(all(doc["op"] == "n" for doc in self.opman.init_cursor()[0]))
        self.assertEqual(self.opman.checkpoint, None)

        # No last checkpoint, empty collections, nothing in oplog
        self.opman.collection_dump = True
        cursor, cursor_empty = self.opman.init_cursor()
        self.assertEqual(cursor, None)
        self.assertTrue(cursor_empty)
        self.assertEqual(self.opman.checkpoint, None)

        # No last checkpoint, empty collections, something in oplog
        self.opman.oplog = self.primary_conn["local"]["oplog.rs"]
        collection = self.primary_conn["includedb1"]["includecol1"]
        collection.insert_one({"idb1col1": 1})
        collection.delete_one({"idb1col1": 1})
        time.sleep(3)
        last_ts = self.opman.get_last_oplog_timestamp()
        cursor, cursor_empty = self.opman.init_cursor()
        self.assertFalse(cursor_empty)
        self.assertEqual(self.opman.checkpoint, last_ts)
        self.assertEqual(self.opman.read_last_checkpoint(), last_ts)

        # No last checkpoint, no collection dump, something in oplog
        # If collection dump is false the checkpoint should not be set
        self.opman.checkpoint = None
        self.opman.oplog_progress = LockingDict()
        self.opman.collection_dump = False
        collection.insert_one({"idb1col1": 2})
        cursor, cursor_empty = self.opman.init_cursor()
        for doc in cursor:
            last_doc = doc
        self.assertEqual(last_doc["o"]["idb1col1"], 2)
        self.assertIsNone(self.opman.checkpoint)

        # Last checkpoint exists
        collection.insert_many([{"idb1col1": i + 500} for i in range(1000)])
        entry = list(self.primary_conn["local"]["oplog.rs"].find(skip=200, limit=-2))
        self.opman.update_checkpoint(entry[0]["ts"])
        cursor, cursor_empty = self.opman.init_cursor()
        self.assertEqual(next(cursor)["ts"], entry[1]["ts"])
        self.assertEqual(self.opman.checkpoint, entry[0]["ts"])
        self.assertEqual(self.opman.read_last_checkpoint(), entry[0]["ts"])

        # Last checkpoint is behind
        self.opman.update_checkpoint(bson.Timestamp(1, 0))
        cursor, cursor_empty = self.opman.init_cursor()
        self.assertTrue(cursor_empty)
        self.assertEqual(cursor, None)
        self.assertEqual(self.opman.checkpoint, bson.Timestamp(1, 0))

    def test_namespace_mapping(self):
        """Test mapping of namespaces
        Cases:

        upsert/delete/update of documents:
        1. in namespace set, mapping provided
        2. outside of namespace set, mapping provided
        """
        included_names = [
            "includedb1.includecol1",
            "includedb1.includecol2",
            "includedb2.includecol1",
        ]
        excluded_names = ["includedb2.excludecol2", "excludedb3.excludecol1"]
        dest_mapping = {
            "includedb1.*": "newdb1.*",
            "includedb2.includecol1": "newdb2.newcol1",
        }
        self.reset_opman(dest_mapping=dest_mapping)
        docman = self.opman.doc_managers[0]
        namespace_config = self.opman.namespace_config

        # start replicating
        self.opman.start()

        base_doc = {"_id": 1, "name": "superman"}

        for ns in included_names:
            db, coll = ns.split(".", 1)

            # test insert
            self.primary_conn[db][coll].insert_one(base_doc)

            assert_soon(lambda: len(docman._search()) == 1)
            self.assertEqual(
                docman._search()[0]["ns"], namespace_config.map_namespace(ns)
            )
            bad = [d for d in docman._search() if d["ns"] == ns]
            self.assertEqual(len(bad), 0)

            # test update
            self.primary_conn[db][coll].update_one(
                {"_id": 1}, {"$set": {"weakness": "kryptonite"}}
            )

            def update_complete():
                docs = docman._search()
                for d in docs:
                    if d.get("weakness") == "kryptonite":
                        return True
                    return False

            assert_soon(update_complete)
            self.assertEqual(
                docman._search()[0]["ns"], namespace_config.map_namespace(ns)
            )
            bad = [d for d in docman._search() if d["ns"] == ns]
            self.assertEqual(len(bad), 0)

            # test delete
            self.primary_conn[db][coll].delete_one({"_id": 1})
            assert_soon(lambda: len(docman._search()) == 0)
            bad = [
                d
                for d in docman._search()
                if d["ns"] == namespace_config.map_namespace(ns)
            ]
            self.assertEqual(len(bad), 0)

            # cleanup
            self.primary_conn[db][coll].drop()
            self.opman.doc_managers[0]._delete()

        # doc not in namespace set
        for ns in excluded_names:
            db, coll = ns.split(".", 1)

            # test insert
            self.primary_conn[db][coll].insert_one(base_doc)
            time.sleep(1)
            self.assertEqual(len(docman._search()), 0)
            # test update
            self.primary_conn[db][coll].update_one(
                {"_id": 1}, {"$set": {"weakness": "kryptonite"}}
            )
            time.sleep(1)
            self.assertEqual(len(docman._search()), 0)

    def test_many_targets(self):
        """Test that one OplogThread is capable of replicating to more than
        one target.
        """

        self.reset_opman(["includedb1.*"], [], {})
        doc_managers = [DocManager(), DocManager(), DocManager()]
        self.opman.doc_managers = doc_managers

        # start replicating
        self.opman.start()
        self.primary_conn["includedb1"]["includecol1"].insert_one(
            {"name": "kermit", "color": "green"}
        )
        self.primary_conn["includedb1"]["includecol2"].insert_one(
            {"name": "elmo", "color": "firetruck red"}
        )
        self.primary_conn["excludedb2"]["excludecol1"].insert_one(
            {"name": "panda", "color": "white and black"}
        )

        assert_soon(
            lambda: sum(len(d._search()) for d in doc_managers) == 6,
            "OplogThread should be able to replicate to multiple targets",
        )

        self.primary_conn["includedb1"]["includecol2"].delete_one({"name": "elmo"})

        assert_soon(
            lambda: sum(len(d._search()) for d in doc_managers) == 3,
            "OplogThread should be able to replicate to multiple targets",
        )
        for d in doc_managers:
            self.assertEqual(d._search()[0]["name"], "kermit")


if __name__ == "__main__":
    unittest.main()



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import re
import ssl
import sys

import pymongo

sys.path[0:0] = [""]  # noqa

from mongo_connector import config, errors, connector
from mongo_connector.connector import get_config_options, setup_logging
from mongo_connector.doc_managers import doc_manager_simulator

from tests import unittest


def from_here(*paths):
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), *paths)


class TestConfig(unittest.TestCase):
    """Test parsing a JSON config file into a Config object."""

    def setUp(self):
        self.reset_config()

    def reset_config(self):
        self.options = get_config_options()
        self.conf = config.Config(self.options)

    def load_json(self, d, validate=True, reset_config=True):
        if reset_config:
            self.reset_config()
        # Serialize a python dictionary to json, then load it
        text = json.dumps(d)
        self.conf.load_json(text)
        if validate:
            self.load_options(reset_config=False)

    def load_options(self, d={}, reset_config=True):
        if reset_config:
            self.reset_config()
        argv = []
        for k, v in d.items():
            argv.append(str(k))
            if v is not None:
                argv.append(str(v))
        self.conf.parse_args(argv)

    def test_default(self):
        # Make sure default configuration doesn't raise any exceptions
        self.load_options()

    def test_parse_json(self):
        # Test for basic json parsing correctness
        test_config = {
            "mainAddress": u"testMainAddress",
            "oplogFile": u"testOplogFile",
            "noDump": True,
            "batchSize": 69,
            "verbosity": 3,
            "logging": {
                "type": u"file",
                "filename": u"testFilename",
                "format": u"%(asctime)s [%(levelname)s] %(name)s:%(lineno)d"
                u" - %(message)s",
                "rotationWhen": u"midnight",
                "rotationInterval": 1,
                "rotationBackups": 7,
                "host": u"testHost",
                "facility": u"testFacility",
            },
            "authentication": {
                "adminUsername": u"testAdminUsername",
                "password": u"testPassword",
                "passwordFile": u"testPasswordFile",
            },
            "fields": [u"testFields1", u"testField2"],
            "namespaces": {
                "include": [u"testNamespaceSet"],
                "exclude": [u"testExcludeNamespaceSet"],
                "mapping": {"testMapKey": u"testMapValue"},
                "gridfs": [u"testGridfsSet"],
            },
        }
        self.load_json(test_config, validate=False)

        for test_key in test_config:
            self.assertEqual(self.conf[test_key], test_config[test_key])

        # Test for partial dict updates
        test_config = {
            "logging": {"type": "syslog", "host": "testHost2"},
            "authentication": {
                "adminUsername": "testAdminUsername2",
                "passwordFile": "testPasswordFile2",
            },
            "namespaces": {"exclude": [], "mapping": {}},
        }
        self.load_json(test_config, validate=False, reset_config=False)
        self.assertEqual(
            self.conf["logging"],
            {
                "type": u"syslog",
                "filename": u"testFilename",
                "format": u"%(asctime)s [%(levelname)s] %(name)s:%(lineno)d"
                u" - %(message)s",
                "rotationWhen": u"midnight",
                "rotationInterval": 1,
                "rotationBackups": 7,
                "host": u"testHost2",
                "facility": u"testFacility",
            },
        )
        self.assertEqual(
            self.conf["authentication"],
            {
                "adminUsername": u"testAdminUsername2",
                "password": u"testPassword",
                "passwordFile": u"testPasswordFile2",
            },
        )
        self.assertEqual(
            self.conf["namespaces"],
            {
                "include": [u"testNamespaceSet"],
                "exclude": [],
                "mapping": {},
                "gridfs": [u"testGridfsSet"],
            },
        )

    def test_basic_options(self):
        # Test the assignment of individual options
        def test_option(arg_name, json_key, value, append_cli=True):
            self.load_options({arg_name: value if append_cli else None})
            self.assertEqual(self.conf[json_key], value)

        test_option("-m", "mainAddress", "testMainAddress")
        test_option("-o", "oplogFile", "testOplogFileShort")
        test_option("--batch-size", "batchSize", 69)
        test_option("--continue-on-error", "continueOnError", True, append_cli=False)
        test_option("-v", "verbosity", 3, append_cli=False)

        self.load_options({"-w": "logFile"})
        self.assertEqual(self.conf["logging.type"], "file")
        self.assertEqual(self.conf["logging.filename"], os.path.abspath("logFile"))

        self.load_options(
            {
                "-s": None,
                "--syslog-host": "testHost",
                "--syslog-facility": "testFacility",
            }
        )
        self.assertEqual(self.conf["logging.type"], "syslog")
        self.assertEqual(self.conf["logging.host"], "testHost")
        self.assertEqual(self.conf["logging.facility"], "testFacility")

        self.load_options({"-i": "a,b,c"})
        self.assertEqual(self.conf["fields"], ["a", "b", "c"])

    def test_extraneous_command_line_options(self):
        self.assertRaises(errors.InvalidConfiguration, self.load_options, {"-v": 3})
        # No error.
        self.load_options({"-v": None})

    def test_namespace_set(self):
        # test namespace_set and dest_namespace_set
        self.load_options(
            {
                "-n": "source_db_1.col,source_db_2.col,source_db_3.col",
                "-g": "dest_db_1.col,dest_db_2.col,dest_db_3.col",
            }
        )
        self.assertEqual(
            self.conf["namespaces.include"],
            ["source_db_1.col", "source_db_2.col", "source_db_3.col"],
        )
        self.assertEqual(
            self.conf["namespaces.mapping"],
            {
                "source_db_1.col": "dest_db_1.col",
                "source_db_2.col": "dest_db_2.col",
                "source_db_3.col": "dest_db_3.col",
            },
        )

        # test exclude_namespace_set
        self.load_options({"-x": "source_db_1.col,source_db_2.col,source_db_3.col"})
        self.assertEqual(
            self.conf["namespaces.exclude"],
            ["source_db_1.col", "source_db_2.col", "source_db_3.col"],
        )

    def test_namespace_set_validation(self):
        # duplicate ns_set
        args = {"-n": "a.x,a.x,b.y", "-g": "1.0,2.0,3.0"}
        self.assertRaises(errors.InvalidConfiguration, self.load_options, args)
        d = {"namespaces": {"include": ["a.x", "a.x", "b.y"]}}
        self.assertRaises(errors.InvalidConfiguration, self.load_json, d)

        # duplicate ex_ns_set
        args = {"-x": "a.x,a.x,b.y"}
        self.assertRaises(errors.InvalidConfiguration, self.load_options, args)
        d = {"namespaces": {"exclude": ["a.x", "a.x", "b.y"]}}
        self.assertRaises(errors.InvalidConfiguration, self.load_json, d)

        # duplicate gridfs_set
        args = {"--gridfs-set": "a.x,a.x,b.y"}
        self.assertRaises(errors.InvalidConfiguration, self.load_options, args)
        d = {"namespaces": {"gridfs": ["a.x", "a.x", "b.y"]}}
        self.assertRaises(errors.InvalidConfiguration, self.load_json, d)

        # duplicate dest_ns_set
        args = {"-n": "a.x,b.y,c.z", "--dest-namespace-set": "1.0,3.0,3.0"}
        self.assertRaises(errors.InvalidConfiguration, self.load_options, args)
        d = {"namespaces": {"mapping": {"a.x": "c.z", "b.y": "c.z"}}}
        self.assertRaises(errors.InvalidConfiguration, self.load_json, d)

        # len(ns_set) < len(dest_ns_set)
        args = {"--namespace-set": "a.x,b.y,c.z", "-g": "1.0,2.0,3.0,4.0"}
        self.assertRaises(errors.InvalidConfiguration, self.load_options, args)

        # len(ns_set) > len(dest_ns_set)
        args = {
            "--namespace-set": "a.x,b.y,c.z,d.j",
            "--dest-namespace-set": "1.0,2.0,3.0",
        }
        self.assertRaises(errors.InvalidConfiguration, self.load_options, args)

        # validate ns_set format
        args = {"-n": "a*.x*"}
        self.assertRaises(errors.InvalidConfiguration, self.load_options, args)
        d = {"namespaces": {"include": ["a*.x*"]}}
        self.assertRaises(errors.InvalidConfiguration, self.load_json, d)

        # validate dest_ns_set format
        args = {"-n": "a.x*", "-g": "1*.0*"}
        self.assertRaises(errors.InvalidConfiguration, self.load_options, args)
        d = {"namespaces": {"mapping": {"a*.x*": "1.0"}}}
        self.assertRaises(errors.InvalidConfiguration, self.load_json, d)
        d = {"namespaces": {"mapping": {"a.x*": "1*.0*"}}}
        self.assertRaises(errors.InvalidConfiguration, self.load_json, d)

    def test_validate_mixed_namespace_format(self):
        # It is invalid to combine new and old namespace formats
        mix_namespaces = [
            {"mapping": {"old.format": "old.format2"}, "new.format": True},
            {"gridfs": ["old.format"], "new.format": True},
            {"include": ["old.format"], "new.format": True},
            {"exclude": ["old.format"], "new.format": True},
        ]
        for namespaces in mix_namespaces:
            with self.assertRaises(errors.InvalidConfiguration):
                self.load_json({"namespaces": namespaces})

    def test_doc_managers_from_args(self):
        # Test basic docmanager construction from args
        args = {
            "-d": "doc_manager_simulator",
            "-t": "test_target_url",
            "-u": "id",
            "--auto-commit-interval": 10,
        }
        self.load_options(args)
        self.assertEqual(len(self.conf["docManagers"]), 1)

        dm = self.conf["docManagers"][0]
        self.assertTrue(isinstance(dm, doc_manager_simulator.DocManager))
        self.assertEqual(dm.url, "test_target_url")
        self.assertEqual(dm.unique_key, "id")
        self.assertEqual(dm.auto_commit_interval, 10)

        # no doc_manager but target_url
        args = {"-t": "1,2"}
        self.assertRaises(errors.InvalidConfiguration, self.load_options, args)

    def test_config_validation(self):
        # can't log both to syslog and to logfile
        self.assertRaises(
            errors.InvalidConfiguration,
            self.load_options,
            {"-w": "logFile", "-s": "true"},
        )

        # Can't specify --stdout and logfile
        self.assertRaises(
            errors.InvalidConfiguration,
            self.load_options,
            {"--stdout": None, "-w": "logFile"},
        )

        # can't specify a username without a password
        self.assertRaises(
            errors.InvalidConfiguration, self.load_options, {"-a": "username"}
        )

        # can't specify password and password file
        self.assertRaises(
            errors.InvalidConfiguration,
            self.load_options,
            {"-a": "username", "-p": "password", "-f": "password_file"},
        )

        # docManagers must be a list
        test_config = {"docManagers": "hello"}
        self.assertRaises(errors.InvalidConfiguration, self.load_json, test_config)

        # every element of docManagers must contain a 'docManager' property
        test_config = {"docManagers": [{"targetURL": "testTargetURL"}]}
        self.assertRaises(errors.InvalidConfiguration, self.load_json, test_config)

        # auto commit interval can't be negative
        test_config = {
            "docManagers": [{"docManager": "testDocManager", "autoCommitInterval": -1}]
        }
        self.assertRaises(errors.InvalidConfiguration, self.load_json, test_config)

    def test_ssl_validation(self):
        """Test setting sslCertificatePolicy."""
        # Setting sslCertificatePolicy to not 'ignored' without a CA file
        # PyMongo will attempt to load system provided CA certificates.
        for ssl_cert_req in ["required", "optional"]:
            no_ca_config = {"ssl": {"sslCertificatePolicy": ssl_cert_req}}
            if pymongo.version_tuple < (3, 0):
                self.assertRaises(
                    errors.InvalidConfiguration, self.load_json, no_ca_config
                )
            else:
                self.load_json(no_ca_config)
        # Setting sslCertificatePolicy to an invalid option
        self.assertRaises(
            errors.InvalidConfiguration,
            self.load_json,
            {"ssl": {"sslCACerts": "ca.pem", "sslCertificatePolicy": "invalid"}},
        )


class TestConnectorConfig(unittest.TestCase):
    """Test creating a Connector from a Config."""

    # Configuration where every option is set to a non-default value.
    set_everything_config = {
        "mainAddress": "localhost:12345",
        "oplogFile": from_here("lib", "dummy.timestamp"),
        "noDump": True,
        "batchSize": 3,
        "verbosity": 1,
        "continueOnError": True,
        "timezoneAware": True,
        "logging": {
            "type": "file",
            "filename": from_here("lib", "dummy-connector.log"),
            "rotationWhen": "H",
            "rotationInterval": 3,
            "rotationBackups": 10,
        },
        "authentication": {
            "adminUsername": "elmo",
            "passwordFile": from_here("lib", "dummy.pwd"),
        },
        "ssl": {
            "sslCertfile": "certfile.pem",
            "sslKeyfile": "certfile.key",
            "sslCACerts": "ca.pem",
            "sslCertificatePolicy": "optional",
        },
        "fields": ["field1", "field2", "field3"],
        "namespaces": {
            "include": ["db.source1", "db.source2"],
            "mapping": {"db.source1": "db.dest1", "db.source2": "db.dest2"},
            "gridfs": ["db.fs"],
        },
        "docManagers": [
            {
                "docManager": "doc_manager_simulator",
                "targetURL": "localhost:12345",
                "bulkSize": 500,
                "uniqueKey": "id",
                "autoCommitInterval": 10,
                "args": {"key": "value", "clientOptions": {"foo": "bar"}},
            }
        ],
    }

    # Argv that sets all possible options to a different value from the
    # config JSON above. Some options cannot be reset, since they conflict
    # with the JSON config and will cause an Exception to be raised.
    # Conflicted options are already tested in the TestConfig TestCase.
    set_everything_differently_argv = [
        "-m",
        "localhost:1000",
        "-o",
        from_here("lib", "bar.timestamp"),
        "--batch-size",
        "100",
        "--verbose",
        "--logfile-when",
        "D",
        "--logfile-interval",
        "5",
        "--logfile-backups",
        "10",
        "--fields",
        "fieldA,fieldB",
        "--gridfs-set",
        "db.gridfs",
        "--unique-key",
        "customer_id",
        "--auto-commit-interval",
        "100",
        "--continue-on-error",
        "-t",
        "localhost:54321",
        "-d",
        "doc_manager_simulator",
        "-n",
        "foo.bar,fiz.biz",
        "-g",
        "foo2.bar2,fiz2.biz2",
        "--ssl-certfile",
        "certfile2.pem",
        "--ssl-ca-certs",
        "ca2.pem",
        "--ssl-certificate-policy",
        "ignored",
    ]

    # Set of files to keep in the 'lib' directory after each run.
    # The Connector and OplogThread create their own files in this directory
    # that should be cleaned out between tests.
    files_to_keep = set(("dummy.pwd",))

    def setUp(self):
        self.config = config.Config(get_config_options())
        # Remove all logging Handlers, since tests may create Handlers.
        logger = logging.getLogger()
        for handler in logger.handlers:
            logger.removeHandler(handler)

    def tearDown(self):
        for filename in os.listdir(from_here("lib")):
            if filename not in self.files_to_keep:
                try:
                    os.remove(from_here("lib", filename))
                except OSError:
                    pass  # File may no longer exist.

    def assertConnectorState(self):
        """Assert that a Connector is constructed from a Config properly."""
        mc = connector.Connector.from_config(self.config)

        # Test Connector options.
        self.assertEqual(mc.address, self.config["mainAddress"])
        self.assertIsInstance(mc.doc_managers[0], doc_manager_simulator.DocManager)

        pwfile = self.config["authentication.passwordFile"]
        if pwfile:
            with open(pwfile, "r") as fd:
                test_password = re.sub(r"\s", "", fd.read())
                self.assertEqual(mc.auth_key, test_password)

        self.assertEqual(mc.auth_username, self.config["authentication.adminUsername"])
        self.assertEqual(mc.oplog_checkpoint, self.config["oplogFile"])
        self.assertEqual(mc.tz_aware, self.config["timezoneAware"])
        self.assertEqual(
            mc.ssl_kwargs.get("ssl_certfile"), self.config["ssl.sslCertfile"]
        )
        self.assertEqual(
            mc.ssl_kwargs.get("ssl_ca_certs"), self.config["ssl.sslCACerts"]
        )
        self.assertEqual(
            mc.ssl_kwargs.get("ssl_keyfile"), self.config["ssl.sslKeyfile"]
        )
        self.assertEqual(
            mc.ssl_kwargs.get("ssl_cert_reqs"), self.config["ssl.sslCertificatePolicy"]
        )
        command_helper = mc.doc_managers[0].command_helper
        for name in self.config["namespaces.mapping"]:
            self.assertTrue(command_helper.namespace_config.map_namespace(name))

        # Test Logger options.
        log_levels = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
        test_logger = setup_logging(self.config)
        self.assertEqual(log_levels[self.config["verbosity"]], test_logger.level)
        test_handlers = [
            h
            for h in test_logger.handlers
            if isinstance(h, logging.handlers.TimedRotatingFileHandler)
        ]
        self.assertEqual(len(test_handlers), 1)
        test_handler = test_handlers[0]
        expected_handler = logging.handlers.TimedRotatingFileHandler(
            "test-dummy.log",
            when=self.config["logging.rotationWhen"],
            interval=self.config["logging.rotationInterval"],
            backupCount=self.config["logging.rotationBackups"],
        )
        self.assertEqual(test_handler.when, expected_handler.when)
        self.assertEqual(test_handler.backupCount, expected_handler.backupCount)
        self.assertEqual(test_handler.interval, expected_handler.interval)

        # Test keyword arguments passed to OplogThread.
        ot_kwargs = mc.kwargs
        self.assertEqual(ot_kwargs["ns_set"], self.config["namespaces.include"])
        self.assertEqual(ot_kwargs["collection_dump"], not self.config["noDump"])
        self.assertEqual(ot_kwargs["gridfs_set"], self.config["namespaces.gridfs"])
        self.assertEqual(ot_kwargs["continue_on_error"], self.config["continueOnError"])
        self.assertEqual(ot_kwargs["fields"], self.config["fields"])
        self.assertEqual(ot_kwargs["batch_size"], self.config["batchSize"])

        # Test DocManager options.
        for dm, dm_expected in zip(mc.doc_managers, self.config["docManagers"]):
            self.assertEqual(dm.kwargs, dm_expected.kwargs)
            self.assertEqual(dm.auto_commit_interval, dm_expected.auto_commit_interval)
            self.assertEqual(dm.url, dm_expected.url)
            self.assertEqual(dm.chunk_size, dm_expected.chunk_size)

    def test_connector_config_file_options(self):
        # Test Config with only a configuration file.
        self.config.load_json(json.dumps(self.set_everything_config))
        self.config.parse_args(argv=[])
        self.assertConnectorState()

    def test_connector_with_argv(self):
        # Test Config with arguments given on the command-line.
        self.config.parse_args(self.set_everything_differently_argv)
        self.assertConnectorState()

    def test_override_config_with_argv(self):
        # Override some options in the config file with given command-line
        # options.
        self.config.load_json(json.dumps(TestConnectorConfig.set_everything_config))
        self.config.parse_args(self.set_everything_differently_argv)

        first_dm = self.config["docManagers"][0]
        first_dm_config = self.set_everything_config["docManagers"][0]
        self.assertEqual(first_dm.url, "localhost:54321")
        self.assertEqual(first_dm.chunk_size, first_dm_config["bulkSize"])
        self.assertEqual(
            first_dm.kwargs.get("clientOptions"),
            first_dm_config["args"]["clientOptions"],
        )
        self.assertConnectorState()

    def test_client_options(self):
        config_def = {
            "mainAddress": "localhost:27017",
            "oplogFile": from_here("lib", "dummy.timestamp"),
            "docManagers": [
                {
                    "docManager": "mongo_doc_manager",
                    "targetURL": "dummyhost:27017",
                    "args": {"clientOptions": {"maxPoolSize": 50, "connect": False}},
                }
            ],
        }
        config_obj = config.Config(get_config_options())
        config_obj.load_json(json.dumps(config_def))
        config_obj.parse_args(argv=[])
        conn = connector.Connector.from_config(config_obj)
        self.assertEqual(50, conn.doc_managers[0].mongo.max_pool_size)

    def test_ssl_options(self):
        config_def = {
            "mainAddress": "localhost:27017",
            "oplogFile": from_here("lib", "dummy.timestamp"),
            "ssl": {
                "sslCertfile": "certfile.pem",
                "sslKeyfile": "certfile.key",
                "sslCACerts": "ca.pem",
            },
        }
        for cert_policy, expected_ssl_cert_req in [
            ("ignored", ssl.CERT_NONE),
            ("optional", ssl.CERT_OPTIONAL),
            ("required", ssl.CERT_REQUIRED),
            (None, None),
        ]:
            config_def["ssl"]["sslCertificatePolicy"] = cert_policy
            config_obj = config.Config(get_config_options())
            config_obj.load_json(json.dumps(config_def))
            config_obj.parse_args(argv=[])
            mc = connector.Connector.from_config(config_obj)
            self.assertEqual("certfile.pem", mc.ssl_kwargs.get("ssl_certfile"))
            self.assertEqual("ca.pem", mc.ssl_kwargs.get("ssl_ca_certs"))
            self.assertEqual("certfile.key", mc.ssl_kwargs.get("ssl_keyfile"))
            self.assertEqual(expected_ssl_cert_req, mc.ssl_kwargs.get("ssl_cert_reqs"))


if __name__ == "__main__":
    unittest.main()



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test oplog manager methods
"""

import itertools
import re
import sys
import time

import bson
import gridfs
import pymongo

sys.path[0:0] = [""]  # noqa

from mongo_connector.doc_managers.doc_manager_simulator import DocManager
from mongo_connector.locking_dict import LockingDict
from mongo_connector.namespace_config import NamespaceConfig
from mongo_connector.oplog_manager import OplogThread
from mongo_connector.test_utils import assert_soon, close_client, ReplicaSetSingle
from mongo_connector.util import bson_ts_to_long
from tests import unittest


class TestOplogManager(unittest.TestCase):
    """Defines all the testing methods, as well as a method that sets up the
        cluster
    """

    def setUp(self):
        self.repl_set = ReplicaSetSingle().start()
        self.primary_conn = self.repl_set.client()
        self.oplog_coll = self.primary_conn.local["oplog.rs"]
        self.opman = OplogThread(
            primary_client=self.primary_conn,
            doc_managers=(DocManager(),),
            oplog_progress_dict=LockingDict(),
            namespace_config=NamespaceConfig(
                namespace_options={"test.*": True, "gridfs.*": {"gridfs": True}}
            ),
        )

    def tearDown(self):
        try:
            self.opman.join()
        except RuntimeError:
            pass  # OplogThread may not have been started
        self.primary_conn.drop_database("test")
        close_client(self.primary_conn)
        self.repl_set.stop()

    def test_get_oplog_cursor(self):
        """Test the get_oplog_cursor method"""

        # timestamp is None - all oplog entries excluding no-ops are returned.
        cursor = self.opman.get_oplog_cursor(None)
        self.assertEqual(
            cursor.count(),
            self.primary_conn["local"]["oplog.rs"].find({"op": {"$ne": "n"}}).count(),
        )

        # earliest entry is the only one at/after timestamp
        doc = {"ts": bson.Timestamp(1000, 0), "i": 1}
        self.primary_conn["test"]["test"].insert_one(doc)
        latest_timestamp = self.opman.get_last_oplog_timestamp()
        cursor = self.opman.get_oplog_cursor(latest_timestamp)
        self.assertNotEqual(cursor, None)
        self.assertEqual(cursor.count(), 1)
        next_entry_id = next(cursor)["o"]["_id"]
        retrieved = self.primary_conn.test.test.find_one(next_entry_id)
        self.assertEqual(retrieved, doc)

        # many entries before and after timestamp
        self.primary_conn["test"]["test"].insert_many(
            [{"i": i} for i in range(2, 1002)]
        )
        oplog_cursor = self.oplog_coll.find(
            {"op": {"$ne": "n"}, "ns": {"$not": re.compile(r"\.(system|\$cmd)")}},
            sort=[("ts", pymongo.ASCENDING)],
        )

        # initial insert + 1000 more inserts
        self.assertEqual(oplog_cursor.count(), 1 + 1000)
        pivot = oplog_cursor.skip(400).limit(-1)[0]

        goc_cursor = self.opman.get_oplog_cursor(pivot["ts"])
        self.assertEqual(goc_cursor.count(), 1 + 1000 - 400)

    def test_get_last_oplog_timestamp(self):
        """Test the get_last_oplog_timestamp method"""

        # "empty" the oplog
        self.opman.oplog = self.primary_conn["test"]["emptycollection"]
        self.assertEqual(self.opman.get_last_oplog_timestamp(), None)

        # Test non-empty oplog
        self.opman.oplog = self.primary_conn["local"]["oplog.rs"]
        for i in range(1000):
            self.primary_conn["test"]["test"].insert_one({"i": i + 500})
        oplog = self.primary_conn["local"]["oplog.rs"]
        oplog = oplog.find().sort("$natural", pymongo.DESCENDING).limit(-1)[0]
        self.assertEqual(self.opman.get_last_oplog_timestamp(), oplog["ts"])

    def test_dump_collection(self):
        """Test the dump_collection method

        Cases:

        1. empty oplog
        2. non-empty oplog, with gridfs collections
        3. non-empty oplog, specified a namespace-set, none of the oplog
           entries are for collections in the namespace-set
        """

        # Test with empty oplog
        self.opman.oplog = self.primary_conn["test"]["emptycollection"]
        last_ts = self.opman.dump_collection()
        self.assertEqual(last_ts, None)

        # Test with non-empty oplog with gridfs collections
        self.opman.oplog = self.primary_conn["local"]["oplog.rs"]
        # Insert 10 gridfs files
        for i in range(10):
            fs = gridfs.GridFS(self.primary_conn["gridfs"], collection="test" + str(i))
            fs.put(b"hello world")
        # Insert 1000 documents
        for i in range(1000):
            self.primary_conn["test"]["test"].insert_one({"i": i + 500})
        last_ts = self.opman.get_last_oplog_timestamp()
        self.assertEqual(last_ts, self.opman.dump_collection())
        self.assertEqual(len(self.opman.doc_managers[0]._search()), 1010)

        # Case 3
        # 1MB oplog so that we can rollover quickly
        repl_set = ReplicaSetSingle(oplogSize=1).start()
        conn = repl_set.client()
        opman = OplogThread(
            primary_client=conn,
            doc_managers=(DocManager(),),
            oplog_progress_dict=LockingDict(),
            namespace_config=NamespaceConfig(namespace_set=["test.test"]),
        )
        # Insert a document into an included collection
        conn["test"]["test"].insert_one({"test": 1})
        # Cause the oplog to rollover on a non-included collection
        while conn["local"]["oplog.rs"].find_one({"ns": "test.test"}):
            conn["test"]["ignored"].insert_many(
                [{"test": "1" * 1024} for _ in range(1024)]
            )
        last_ts = opman.get_last_oplog_timestamp()
        self.assertEqual(last_ts, opman.dump_collection())
        self.assertEqual(len(opman.doc_managers[0]._search()), 1)
        conn.close()
        repl_set.stop()

    def test_skipped_oplog_entry_updates_checkpoint(self):
        repl_set = ReplicaSetSingle().start()
        conn = repl_set.client()
        opman = OplogThread(
            primary_client=conn,
            doc_managers=(DocManager(),),
            oplog_progress_dict=LockingDict(),
            namespace_config=NamespaceConfig(namespace_set=["test.test"]),
        )
        opman.start()

        # Insert a document into an included collection
        conn["test"]["test"].insert_one({"test": 1})
        last_ts = opman.get_last_oplog_timestamp()
        assert_soon(
            lambda: last_ts == opman.checkpoint,
            "OplogThread never updated checkpoint to non-skipped " "entry.",
        )
        self.assertEqual(len(opman.doc_managers[0]._search()), 1)

        # Make sure that the oplog thread updates its checkpoint on every
        # oplog entry.
        conn["test"]["ignored"].insert_one({"test": 1})
        last_ts = opman.get_last_oplog_timestamp()
        assert_soon(
            lambda: last_ts == opman.checkpoint,
            "OplogThread never updated checkpoint to skipped entry.",
        )
        opman.join()
        conn.close()
        repl_set.stop()

    def test_dump_collection_with_error(self):
        """Test the dump_collection method with invalid documents.

        Cases:

        1. non-empty oplog, continue_on_error=True, invalid documents
        """

        # non-empty oplog, continue_on_error=True, invalid documents
        self.opman.continue_on_error = True
        self.opman.oplog = self.primary_conn["local"]["oplog.rs"]

        docs = [{"a": i} for i in range(100)]
        for i in range(50, 60):
            docs[i]["_upsert_exception"] = True
        self.primary_conn["test"]["test"].insert_many(docs)

        last_ts = self.opman.get_last_oplog_timestamp()
        self.assertEqual(last_ts, self.opman.dump_collection())
        docs = self.opman.doc_managers[0]._search()
        docs.sort(key=lambda doc: doc["a"])

        self.assertEqual(len(docs), 90)
        expected_a = itertools.chain(range(0, 50), range(60, 100))
        for doc, correct_a in zip(docs, expected_a):
            self.assertEqual(doc["a"], correct_a)

    def test_dump_collection_cancel(self):
        """Test that dump_collection returns None when cancelled."""
        self.primary_conn["test"]["test"].insert_one({"test": "1"})

        # Pretend that the OplogThead was cancelled
        self.opman.running = False
        self.assertIsNone(self.opman.dump_collection())

    def test_init_cursor(self):
        """Test the init_cursor method

        Cases:

        1. no last checkpoint, no collection dump
        2. no last checkpoint, collection dump ok and stuff to dump
        3. no last checkpoint, nothing to dump, stuff in oplog
        4. no last checkpoint, nothing to dump, nothing in oplog
        5. no last checkpoint, no collection dump, stuff in oplog
        6. last checkpoint exists
        7. last checkpoint is behind
        """

        # N.B. these sub-cases build off of each other and cannot be re-ordered
        # without side-effects

        # No last checkpoint, no collection dump, nothing in oplog
        # "change oplog collection" to put nothing in oplog
        self.opman.oplog = self.primary_conn["test"]["emptycollection"]
        self.opman.collection_dump = False
        self.assertTrue(all(doc["op"] == "n" for doc in self.opman.init_cursor()[0]))
        self.assertEqual(self.opman.checkpoint, None)

        # No last checkpoint, empty collections, nothing in oplog
        self.opman.collection_dump = True
        cursor, cursor_empty = self.opman.init_cursor()
        self.assertEqual(cursor, None)
        self.assertTrue(cursor_empty)
        self.assertEqual(self.opman.checkpoint, None)

        # No last checkpoint, empty collections, something in oplog
        self.opman.oplog = self.primary_conn["local"]["oplog.rs"]
        collection = self.primary_conn["test"]["test"]
        collection.insert_one({"i": 1})
        collection.delete_one({"i": 1})
        time.sleep(3)
        last_ts = self.opman.get_last_oplog_timestamp()
        cursor, cursor_empty = self.opman.init_cursor()
        self.assertFalse(cursor_empty)
        self.assertEqual(self.opman.checkpoint, last_ts)
        self.assertEqual(self.opman.read_last_checkpoint(), last_ts)

        # No last checkpoint, no collection dump, something in oplog
        # If collection dump is false the checkpoint should not be set
        self.opman.checkpoint = None
        self.opman.oplog_progress = LockingDict()
        self.opman.collection_dump = False
        collection.insert_one({"i": 2})
        cursor, cursor_empty = self.opman.init_cursor()
        for doc in cursor:
            last_doc = doc
        self.assertEqual(last_doc["o"]["i"], 2)
        self.assertIsNone(self.opman.checkpoint)

        # Last checkpoint exists, no collection dump, something in oplog
        collection.insert_many([{"i": i + 500} for i in range(1000)])
        entry = list(self.primary_conn["local"]["oplog.rs"].find(skip=200, limit=-2))
        self.opman.update_checkpoint(entry[0]["ts"])
        cursor, cursor_empty = self.opman.init_cursor()
        self.assertEqual(next(cursor)["ts"], entry[1]["ts"])
        self.assertEqual(self.opman.checkpoint, entry[0]["ts"])
        self.assertEqual(self.opman.read_last_checkpoint(), entry[0]["ts"])

        # Last checkpoint is behind
        self.opman.update_checkpoint(bson.Timestamp(1, 0))
        cursor, cursor_empty = self.opman.init_cursor()
        self.assertTrue(cursor_empty)
        self.assertEqual(cursor, None)
        self.assertEqual(self.opman.checkpoint, bson.Timestamp(1, 0))

    def test_namespace_mapping(self):
        """Test mapping of namespaces
        Cases:

        upsert/delete/update of documents:
        1. in namespace set, mapping provided
        2. outside of namespace set, mapping provided
        """

        source_ns = ["test.test1", "test.test2"]
        phony_ns = ["test.phony1", "test.phony2"]
        dest_mapping = {
            "test.test1": "test.test1_dest",
            "test.test2": "test.test2_dest",
        }
        self.opman.namespace_config = NamespaceConfig(
            namespace_set=source_ns, namespace_options=dest_mapping
        )
        docman = self.opman.doc_managers[0]
        # start replicating
        self.opman.start()

        base_doc = {"_id": 1, "name": "superman"}

        # doc in namespace set
        for ns in source_ns:
            db, coll = ns.split(".", 1)

            # test insert
            self.primary_conn[db][coll].insert_one(base_doc)

            assert_soon(lambda: len(docman._search()) == 1)
            self.assertEqual(docman._search()[0]["ns"], dest_mapping[ns])
            bad = [d for d in docman._search() if d["ns"] == ns]
            self.assertEqual(len(bad), 0)

            # test update
            self.primary_conn[db][coll].update_one(
                {"_id": 1}, {"$set": {"weakness": "kryptonite"}}
            )

            def update_complete():
                docs = docman._search()
                for d in docs:
                    if d.get("weakness") == "kryptonite":
                        return True
                    return False

            assert_soon(update_complete)
            self.assertEqual(docman._search()[0]["ns"], dest_mapping[ns])
            bad = [d for d in docman._search() if d["ns"] == ns]
            self.assertEqual(len(bad), 0)

            # test delete
            self.primary_conn[db][coll].delete_one({"_id": 1})
            assert_soon(lambda: len(docman._search()) == 0)
            bad = [d for d in docman._search() if d["ns"] == dest_mapping[ns]]
            self.assertEqual(len(bad), 0)

            # cleanup
            self.primary_conn[db][coll].delete_many({})
            self.opman.doc_managers[0]._delete()

        # doc not in namespace set
        for ns in phony_ns:
            db, coll = ns.split(".", 1)

            # test insert
            self.primary_conn[db][coll].insert_one(base_doc)
            time.sleep(1)
            self.assertEqual(len(docman._search()), 0)
            # test update
            self.primary_conn[db][coll].update_one(
                {"_id": 1}, {"$set": {"weakness": "kryptonite"}}
            )
            time.sleep(1)
            self.assertEqual(len(docman._search()), 0)

    def test_many_targets(self):
        """Test that one OplogThread is capable of replicating to more than
        one target.
        """
        doc_managers = [DocManager(), DocManager(), DocManager()]
        self.opman.doc_managers = doc_managers

        # start replicating
        self.opman.start()
        self.primary_conn["test"]["test"].insert_one(
            {"name": "kermit", "color": "green"}
        )
        self.primary_conn["test"]["test"].insert_one(
            {"name": "elmo", "color": "firetruck red"}
        )

        assert_soon(
            lambda: sum(len(d._search()) for d in doc_managers) == 6,
            "OplogThread should be able to replicate to multiple targets",
        )

        self.primary_conn["test"]["test"].delete_one({"name": "elmo"})

        assert_soon(
            lambda: sum(len(d._search()) for d in doc_managers) == 3,
            "OplogThread should be able to replicate to multiple targets",
        )
        for d in doc_managers:
            self.assertEqual(d._search()[0]["name"], "kermit")

    def test_upgrade_oplog_progress(self):
        first_oplog_ts = self.opman.oplog.find_one()["ts"]
        # Old format oplog progress file:
        progress = {str(self.opman.oplog): bson_ts_to_long(first_oplog_ts)}
        # Set up oplog managers to use the old format.
        oplog_progress = LockingDict()
        oplog_progress.dict = progress
        self.opman.oplog_progress = oplog_progress
        # Cause the oplog managers to update their checkpoints.
        self.opman.update_checkpoint(first_oplog_ts)
        # New format should be in place now.
        new_format = {self.opman.replset_name: first_oplog_ts}
        self.assertEqual(new_format, self.opman.oplog_progress.get_dict())


if __name__ == "__main__":
    unittest.main()



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import importlib_metadata


__version__ = importlib_metadata.version("mongo_connector")


# Maximum # of documents to process before recording timestamp
# default = -1 (no maximum)
DEFAULT_BATCH_SIZE = -1

# Interval in seconds between doc manager flushes (i.e. auto commit)
# default = None (never auto commit)
DEFAULT_COMMIT_INTERVAL = None

# Maximum # of documents to send in a single bulk request through a
# DocManager.
DEFAULT_MAX_BULK = 1000

# The default MongoDB field that will serve as the unique key for the
# target system.
DEFAULT_UNIQUE_KEY = "_id"

# Default host and facility for logging to the syslog.
DEFAULT_SYSLOG_HOST = "localhost:514"
DEFAULT_SYSLOG_FACILITY = "user"

# ROTATING LOGFILE
# The type of interval
# (seconds, minutes, hours... c.f. logging.handlers.TimedRotatingFileHandler)
DEFAULT_LOGFILE_WHEN = "midnight"
# The rollover interval
DEFAULT_LOGFILE_INTERVAL = 1
# Number of log files to keep
DEFAULT_LOGFILE_BACKUPCOUNT = 7
# The log format
DEFAULT_LOG_FORMAT = "%(asctime)s [%(levelname)s] %(name)s:%(lineno)d - %(message)s"

# If a single meta collection is used, defines the default collection name
DEFAULT_META_COLLECTION_NAME = "__oplog"
# If a single meta collection is used, defines the default cap size
DEFAULT_META_COLLECTION_CAP_SIZE = 5 * 1024 * 1024



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Discovers the MongoDB cluster and starts the connector.
"""

import copy
import json
import logging
import logging.handlers
import os
import platform
import pymongo
import re
import shutil
import signal
import ssl
import sys
import threading
import time

from pymongo import MongoClient

from mongo_connector import config, constants, errors, util
from mongo_connector.constants import __version__
from mongo_connector.locking_dict import LockingDict
from mongo_connector.oplog_manager import OplogThread
from mongo_connector.command_helper import CommandHelper
from mongo_connector.util import log_fatal_exceptions, retry_until_ok
from mongo_connector.namespace_config import NamespaceConfig, validate_namespace_options
from mongo_connector.version import Version


# Monkey patch logging to add Logger.always
ALWAYS = logging.CRITICAL + 10
logging.addLevelName(ALWAYS, "ALWAYS")


def always(self, message, *args, **kwargs):
    self.log(ALWAYS, message, *args, **kwargs)


logging.Logger.always = always


LOG = logging.getLogger(__name__)

_SSL_POLICY_MAP = {
    "ignored": ssl.CERT_NONE,
    "optional": ssl.CERT_OPTIONAL,
    "required": ssl.CERT_REQUIRED,
}

_mininum_mongodb_version = None
"""The minimum MongoDB version in the source cluster."""


def get_mininum_mongodb_version():
    return _mininum_mongodb_version


def update_mininum_mongodb_version(version):
    global _mininum_mongodb_version
    if version is None:
        _mininum_mongodb_version = version
    if _mininum_mongodb_version is None or version < _mininum_mongodb_version:
        _mininum_mongodb_version = version


class Connector(threading.Thread):
    """Thread that monitors a replica set or sharded cluster.

    Creates, runs, and monitors an OplogThread for each replica set found.
    """

    def __init__(self, mongo_address, doc_managers=None, **kwargs):
        super(Connector, self).__init__()

        # can_run is set to false when we join the thread
        self.can_run = True

        # The signal that caused the connector to stop or None
        self.signal = None

        # main address - either mongos for sharded setups or a primary otherwise
        self.address = mongo_address

        # connection to the main address
        self.main_conn = None

        # List of DocManager instances
        if doc_managers:
            self.doc_managers = doc_managers
        else:
            LOG.warning("No doc managers specified, using simulator.")
            # Avoid circular import on get_mininum_mongodb_version.
            from mongo_connector.doc_managers import doc_manager_simulator

            self.doc_managers = (doc_manager_simulator.DocManager(),)

        # Password for authentication
        self.auth_key = kwargs.pop("auth_key", None)

        # Username for authentication
        self.auth_username = kwargs.pop("auth_username", None)

        # The name of the file that stores the progress of the OplogThreads
        self.oplog_checkpoint = kwargs.pop("oplog_checkpoint", "oplog.timestamp")

        # The set of OplogThreads created
        self.shard_set = {}

        # Dict of OplogThread/timestamp pairs to record progress
        self.oplog_progress = LockingDict()

        # Timezone awareness
        self.tz_aware = kwargs.get("tz_aware", False)

        # SSL keyword arguments to MongoClient.
        ssl_certfile = kwargs.pop("ssl_certfile", None)
        ssl_ca_certs = kwargs.pop("ssl_ca_certs", None)
        ssl_keyfile = kwargs.pop("ssl_keyfile", None)
        ssl_cert_reqs = kwargs.pop("ssl_cert_reqs", None)
        self.ssl_kwargs = {}
        if ssl_certfile is not None:
            self.ssl_kwargs["ssl_certfile"] = ssl_certfile
        if ssl_ca_certs is not None:
            self.ssl_kwargs["ssl_ca_certs"] = ssl_ca_certs
        if ssl_keyfile is not None:
            self.ssl_kwargs["ssl_keyfile"] = ssl_keyfile
        if ssl_cert_reqs is not None:
            self.ssl_kwargs["ssl_cert_reqs"] = ssl_cert_reqs

        # Save the rest of kwargs.
        self.kwargs = kwargs

        # The namespace configuration shared by all OplogThreads and
        # DocManagers
        self.namespace_config = NamespaceConfig(
            namespace_set=kwargs.get("ns_set"),
            ex_namespace_set=kwargs.get("ex_ns_set"),
            gridfs_set=kwargs.get("gridfs_set"),
            dest_mapping=kwargs.get("dest_mapping"),
            namespace_options=kwargs.get("namespace_options"),
            include_fields=kwargs.get("fields"),
            exclude_fields=kwargs.get("exclude_fields"),
        )

        # Initialize and set the command helper
        command_helper = CommandHelper(self.namespace_config)
        for dm in self.doc_managers:
            dm.command_helper = command_helper

        if self.oplog_checkpoint is not None:
            if not os.path.exists(self.oplog_checkpoint):
                info_str = (
                    "MongoConnector: Can't find %s, "
                    "attempting to create an empty progress log" % self.oplog_checkpoint
                )
                LOG.warning(info_str)
                try:
                    # Create oplog progress file
                    open(self.oplog_checkpoint, "w").close()
                except IOError as e:
                    LOG.critical(
                        "MongoConnector: Could not "
                        "create a progress log: %s" % str(e)
                    )
                    sys.exit(2)
            else:
                if not os.access(self.oplog_checkpoint, os.W_OK) and not os.access(
                    self.oplog_checkpoint, os.R_OK
                ):
                    LOG.critical(
                        "Invalid permissions on %s! Exiting" % (self.oplog_checkpoint)
                    )
                    sys.exit(2)

    @classmethod
    def from_config(cls, config):
        """Create a new Connector instance from a Config object."""
        auth_key = None
        password_file = config["authentication.passwordFile"]
        if password_file is not None:
            try:
                auth_key = open(config["authentication.passwordFile"]).read()
                auth_key = re.sub(r"\s", "", auth_key)
            except IOError:
                LOG.error("Could not load password file!")
                sys.exit(1)
        password = config["authentication.password"]
        if password is not None:
            auth_key = password
        connector = Connector(
            mongo_address=config["mainAddress"],
            doc_managers=config["docManagers"],
            oplog_checkpoint=os.path.abspath(config["oplogFile"]),
            collection_dump=config["onlyDump"] or not config["noDump"],
            only_dump=config["onlyDump"],
            batch_size=config["batchSize"],
            continue_on_error=config["continueOnError"],
            auth_username=config["authentication.adminUsername"],
            auth_key=auth_key,
            fields=config["fields"],
            exclude_fields=config["exclude_fields"],
            ns_set=config["namespaces.include"],
            ex_ns_set=config["namespaces.exclude"],
            dest_mapping=config["namespaces.mapping"],
            namespace_options=config["namespaces.namespace_options"],
            gridfs_set=config["namespaces.gridfs"],
            ssl_certfile=config["ssl.sslCertfile"],
            ssl_keyfile=config["ssl.sslKeyfile"],
            ssl_ca_certs=config["ssl.sslCACerts"],
            ssl_cert_reqs=config["ssl.sslCertificatePolicy"],
            tz_aware=config["timezoneAware"],
        )
        return connector

    def join(self):
        """ Joins thread, stops it from running
        """
        self.can_run = False
        super(Connector, self).join()
        for dm in self.doc_managers:
            dm.stop()

    def write_oplog_progress(self):
        """ Writes oplog progress to file provided by user
        """

        if self.oplog_checkpoint is None:
            return None

        with self.oplog_progress as oplog_prog:
            oplog_dict = oplog_prog.get_dict()
        items = [[name, util.bson_ts_to_long(oplog_dict[name])] for name in oplog_dict]
        if not items:
            return

        # write to temp file
        backup_file = self.oplog_checkpoint + ".backup"
        os.rename(self.oplog_checkpoint, backup_file)

        # for each of the threads write to file
        with open(self.oplog_checkpoint, "w") as dest:
            if len(items) == 1:
                # Write 1-dimensional array, as in previous versions.
                json_str = json.dumps(items[0])
            else:
                # Write a 2d array to support sharded clusters.
                json_str = json.dumps(items)
            try:
                dest.write(json_str)
            except IOError:
                # Basically wipe the file, copy from backup
                dest.truncate()
                with open(backup_file, "r") as backup:
                    shutil.copyfile(backup, dest)

        os.remove(backup_file)

    def read_oplog_progress(self):
        """Reads oplog progress from file provided by user.
        This method is only called once before any threads are spanwed.
        """

        if self.oplog_checkpoint is None:
            return None

        # Check for empty file
        try:
            if os.stat(self.oplog_checkpoint).st_size == 0:
                LOG.info("MongoConnector: Empty oplog progress file.")
                return None
        except OSError:
            return None

        with open(self.oplog_checkpoint, "r") as progress_file:
            try:
                data = json.load(progress_file)
            except ValueError:
                LOG.exception(
                    'Cannot read oplog progress file "%s". '
                    "It may be corrupt after Mongo Connector was shut down"
                    "uncleanly. You can try to recover from a backup file "
                    '(may be called "%s.backup") or create a new progress file '
                    "starting at the current moment in time by running "
                    "mongo-connector --no-dump <other options>. "
                    "You may also be trying to read an oplog progress file "
                    "created with the old format for sharded clusters. "
                    "See https://github.com/10gen-labs/mongo-connector/wiki"
                    "/Oplog-Progress-File for complete documentation."
                    % (self.oplog_checkpoint, self.oplog_checkpoint)
                )
                return
            # data format:
            # [name, timestamp] = replica set
            # [[name, timestamp], [name, timestamp], ...] = sharded cluster
            if not isinstance(data[0], list):
                data = [data]
            with self.oplog_progress:
                self.oplog_progress.dict = dict(
                    (name, util.long_to_bson_ts(timestamp)) for name, timestamp in data
                )

    @staticmethod
    def copy_uri_options(hosts, mongodb_uri):
        """Returns a MongoDB URI to hosts with the options from mongodb_uri.
        """
        if "?" in mongodb_uri:
            options = mongodb_uri.split("?", 1)[1]
        else:
            options = None
        uri = "mongodb://" + hosts
        if options:
            uri += "/?" + options
        return uri

    def create_authed_client(self, hosts=None, **kwargs):
        kwargs.update(self.ssl_kwargs)
        if hosts is None:
            new_uri = self.address
        else:
            new_uri = self.copy_uri_options(hosts, self.address)
        client = MongoClient(new_uri, tz_aware=self.tz_aware, **kwargs)
        if self.auth_key is not None:
            client["admin"].authenticate(self.auth_username, self.auth_key)
        return client

    def update_version_from_client(self, client):
        is_master = client.admin.command("isMaster")
        for host in is_master["hosts"]:
            update_mininum_mongodb_version(
                Version.from_client(self.create_authed_client(host))
            )

    @log_fatal_exceptions
    def run(self):
        """Discovers the mongo cluster and creates a thread for each primary.
        """
        # Reset the global minimum MongoDB version
        update_mininum_mongodb_version(None)
        self.main_conn = self.create_authed_client()
        LOG.always(
            "Source MongoDB version: %s",
            self.main_conn.admin.command("buildInfo")["version"],
        )

        for dm in self.doc_managers:
            name = dm.__class__.__module__
            module = sys.modules[name]
            version = "unknown"
            if hasattr(module, "__version__"):
                version = module.__version__
            elif hasattr(module, "version"):
                version = module.version
            LOG.always("Target DocManager: %s version: %s", name, version)

        self.read_oplog_progress()
        conn_type = None

        try:
            self.main_conn.admin.command("isdbgrid")
        except pymongo.errors.OperationFailure:
            conn_type = "REPLSET"

        if conn_type == "REPLSET":
            # Make sure we are connected to a replica set
            is_master = self.main_conn.admin.command("isMaster")
            if "setName" not in is_master:
                LOG.error(
                    'No replica set at "%s"! A replica set is required '
                    "to run mongo-connector. Shutting down..." % self.address
                )
                return

            # Establish a connection to the replica set as a whole
            self.main_conn.close()
            self.main_conn = self.create_authed_client(replicaSet=is_master["setName"])

            self.update_version_from_client(self.main_conn)

            # non sharded configuration
            oplog = OplogThread(
                self.main_conn,
                self.doc_managers,
                self.oplog_progress,
                self.namespace_config,
                **self.kwargs
            )
            self.shard_set[0] = oplog
            LOG.info("MongoConnector: Starting connection thread %s" % self.main_conn)
            oplog.start()

            while self.can_run:
                shard_thread = self.shard_set[0]
                if not (shard_thread.running and shard_thread.is_alive()):
                    LOG.error(
                        "MongoConnector: OplogThread"
                        " %s unexpectedly stopped! Shutting down"
                        % (str(self.shard_set[0]))
                    )
                    self.oplog_thread_join()
                    for dm in self.doc_managers:
                        dm.stop()
                    return

                self.write_oplog_progress()
                time.sleep(1)

        else:  # sharded cluster
            while self.can_run:
                # The backup role does not provide the listShards privilege,
                # so use the config.shards collection instead.
                for shard_doc in retry_until_ok(
                    lambda: list(self.main_conn.config.shards.find())
                ):
                    shard_id = shard_doc["_id"]
                    if shard_id in self.shard_set:
                        shard_thread = self.shard_set[shard_id]
                        if not (shard_thread.running and shard_thread.is_alive()):
                            LOG.error(
                                "MongoConnector: OplogThread "
                                "%s unexpectedly stopped! Shutting "
                                "down" % (str(self.shard_set[shard_id]))
                            )
                            self.oplog_thread_join()
                            for dm in self.doc_managers:
                                dm.stop()
                            return

                        self.write_oplog_progress()
                        time.sleep(1)
                        continue
                    try:
                        repl_set, hosts = shard_doc["host"].split("/")
                    except ValueError:
                        cause = "The system only uses replica sets!"
                        LOG.exception("MongoConnector: %s", cause)
                        self.oplog_thread_join()
                        for dm in self.doc_managers:
                            dm.stop()
                        return

                    shard_conn = self.create_authed_client(hosts, replicaSet=repl_set)
                    self.update_version_from_client(shard_conn)
                    oplog = OplogThread(
                        shard_conn,
                        self.doc_managers,
                        self.oplog_progress,
                        self.namespace_config,
                        mongos_client=self.main_conn,
                        **self.kwargs
                    )
                    self.shard_set[shard_id] = oplog
                    msg = "Starting connection thread"
                    LOG.info("MongoConnector: %s %s" % (msg, shard_conn))
                    oplog.start()

        if self.signal is not None:
            LOG.info("recieved signal %s: shutting down...", self.signal)
        self.oplog_thread_join()
        self.write_oplog_progress()

    def oplog_thread_join(self):
        """Stops all the OplogThreads
        """
        LOG.info("MongoConnector: Stopping all OplogThreads")
        for thread in self.shard_set.values():
            thread.join()


def get_config_options():
    result = []

    def add_option(*args, **kwargs):
        opt = config.Option(*args, **kwargs)
        result.append(opt)
        return opt

    main_address = add_option(
        config_key="mainAddress", default="localhost:27017", type=str
    )

    # -m is for the main address, which is a host:port pair, ideally of the
    # mongos. For non sharded clusters, it can be the primary.
    main_address.add_cli(
        "-m",
        "--main",
        dest="main_address",
        help="Specify the main address, which is a"
        " host:port pair. For sharded clusters, this"
        " should be the mongos address. For individual"
        " replica sets, supply the address of the"
        " primary. For example, `-m localhost:27217`"
        " would be a valid argument to `-m`. Don't use"
        " quotes around the address.",
    )

    oplog_file = add_option(config_key="oplogFile", default="oplog.timestamp", type=str)

    # -o is to specify the oplog-config file. This file is used by the system
    # to store the last timestamp read on a specific oplog. This allows for
    # quick recovery from failure.
    oplog_file.add_cli(
        "-o",
        "--oplog-ts",
        dest="oplog_file",
        help="Specify the name of the file that stores the "
        "oplog progress timestamps. "
        "This file is used by the system to store the last "
        "timestamp read on a specific oplog. This allows "
        "for quick recovery from failure. By default this "
        "is `config.txt`, which starts off empty. An empty "
        "file causes the system to go through all the mongo "
        "oplog and sync all the documents. Whenever the "
        "cluster is restarted, it is essential that the "
        "oplog-timestamp config file be emptied - otherwise "
        "the connector will miss some documents and behave "
        "incorrectly.",
    )

    no_dump = add_option(config_key="noDump", default=False, type=bool)

    # --no-dump specifies whether we should read an entire collection from
    # scratch if no timestamp is found in the oplog_config.
    no_dump.add_cli(
        "--no-dump",
        action="store_true",
        dest="no_dump",
        help="If specified, this flag will ensure that "
        "mongo_connector won't read the entire contents of a "
        "namespace iff --oplog-ts points to an empty file.",
    )

    # --only-dump specifies whether or not we should read the oplog after
    # the initial dump is finished
    only_dump = add_option(
        config_key="onlyDump",
        default=False,
        type=bool)

    only_dump.add_cli(
        "--only-dump", action="store_true", dest="only_dump", help=
        "If specified, this flag will ensure that "
        "mongo_connector won't try to read the oplog after "
        "the namespaces have been dumped.")

    batch_size = add_option(
        config_key="batchSize", default=constants.DEFAULT_BATCH_SIZE, type=int
    )

    # --batch-size specifies num docs to read from oplog before updating the
    # --oplog-ts config file with current oplog position
    batch_size.add_cli(
        "--batch-size",
        type="int",
        dest="batch_size",
        help="Specify an int to update the --oplog-ts "
        "config file with latest position of oplog every "
        "N documents. By default, the oplog config isn't "
        "updated until we've read through the entire oplog. "
        "You may want more frequent updates if you are at risk "
        "of falling behind the earliest timestamp in the oplog",
    )

    def apply_verbosity(option, cli_values):
        if cli_values["verbose"]:
            option.value = 3
        if option.value < 0 or option.value > 3:
            raise errors.InvalidConfiguration("verbosity must be in the range [0, 3].")

    # Default is warnings and above.
    verbosity = add_option(
        config_key="verbosity", default=1, type=int, apply_function=apply_verbosity
    )

    # -v enables verbose logging
    verbosity.add_cli(
        "-v",
        "--verbose",
        action="store_true",
        dest="verbose",
        help="Enables verbose logging.",
    )

    def apply_logging(option, cli_values):
        log_mechs_enabled = [
            cli_values[m]
            for m in ("logfile", "enable_syslog", "stdout")
            if cli_values[m]
        ]
        if len(log_mechs_enabled) > 1:
            raise errors.InvalidConfiguration(
                "You cannot specify more than one logging method "
                "simultaneously. Please choose the logging method you "
                "prefer. "
            )
        if cli_values["log_format"]:
            option.value["format"] = cli_values["log_format"]

        if cli_values["logfile"]:
            when = cli_values["logfile_when"]
            interval = cli_values["logfile_interval"]
            if (
                when
                and when.startswith("W")
                and interval != constants.DEFAULT_LOGFILE_INTERVAL
            ):
                raise errors.InvalidConfiguration(
                    "You cannot specify a log rotation interval when rotating "
                    "based on a weekday (W0 - W6)."
                )

            option.value["type"] = "file"
            option.value["filename"] = cli_values["logfile"]
            if when:
                option.value["rotationWhen"] = when
            if interval:
                option.value["rotationInterval"] = interval
            if cli_values["logfile_backups"]:
                option.value["rotationBackups"] = cli_values["logfile_backups"]

        if cli_values["enable_syslog"]:
            option.value["type"] = "syslog"

        if cli_values["syslog_host"]:
            option.value["host"] = cli_values["syslog_host"]

        if cli_values["syslog_facility"]:
            option.value["facility"] = cli_values["syslog_facility"]

        if cli_values["stdout"]:
            option.value["type"] = "stream"

        # Expand the full path to log file
        option.value["filename"] = os.path.abspath(option.value["filename"])

    default_logging = {
        "type": "file",
        "filename": "mongo-connector.log",
        "format": constants.DEFAULT_LOG_FORMAT,
        "rotationInterval": constants.DEFAULT_LOGFILE_INTERVAL,
        "rotationBackups": constants.DEFAULT_LOGFILE_BACKUPCOUNT,
        "rotationWhen": constants.DEFAULT_LOGFILE_WHEN,
        "host": constants.DEFAULT_SYSLOG_HOST,
        "facility": constants.DEFAULT_SYSLOG_FACILITY,
    }

    logging = add_option(
        config_key="logging",
        default=default_logging,
        type=dict,
        apply_function=apply_logging,
    )

    # -w enables logging to a file
    logging.add_cli(
        "-w", "--logfile", dest="logfile", help="Log all output to the specified file."
    )

    logging.add_cli(
        "--stdout",
        dest="stdout",
        action="store_true",
        help="Log all output to STDOUT rather than a logfile.",
    )

    logging.add_cli(
        "--log-format",
        dest="log_format",
        help="Define a specific format for the log file. "
        "This is based on the python logging lib. "
        "Available parameters can be found at "
        "https://docs.python.org/2/library/logging.html#logrecord-attributes",
    )

    # -s is to enable syslog logging.
    logging.add_cli(
        "-s",
        "--enable-syslog",
        action="store_true",
        dest="enable_syslog",
        help="The syslog host, which may be an address like 'localhost:514' or, "
        "on Unix/Linux, the path to a Unix domain socket such as '/dev/log'.",
    )

    # --syslog-host is to specify the syslog host.
    logging.add_cli(
        "--syslog-host",
        dest="syslog_host",
        help="Used to specify the syslog host." " The default is 'localhost:514'",
    )

    # --syslog-facility is to specify the syslog facility.
    logging.add_cli(
        "--syslog-facility",
        dest="syslog_facility",
        help="Used to specify the syslog facility." " The default is 'user'",
    )

    # --logfile-when specifies the type of interval of the rotating file
    # (seconds, minutes, hours)
    logging.add_cli(
        "--logfile-when",
        action="store",
        dest="logfile_when",
        type="string",
        help="The type of interval for rotating the log file. "
        "Should be one of "
        "'S' (seconds), 'M' (minutes), 'H' (hours), "
        "'D' (days), 'W0' - 'W6' (days of the week 0 - 6), "
        "or 'midnight' (the default). See the Python documentation "
        "for 'logging.handlers.TimedRotatingFileHandler' for more "
        "details.",
    )

    # --logfile-interval specifies when to create a new log file
    logging.add_cli(
        "--logfile-interval",
        action="store",
        dest="logfile_interval",
        type="int",
        help="How frequently to rotate the log file, "
        "specifically, how many units of the rotation interval "
        "should pass before the rotation occurs. For example, "
        "to create a new file each hour: "
        " '--logfile-when=H --logfile-interval=1'. "
        "Defaults to 1. You may not use this option if "
        "--logfile-when is set to a weekday (W0 - W6). "
        "See the Python documentation for "
        "'logging.handlers.TimedRotatingFileHandler' for more "
        "details. ",
    )

    # --logfile-backups specifies how many log files will be kept.
    logging.add_cli(
        "--logfile-backups",
        action="store",
        dest="logfile_backups",
        type="int",
        help="How many log files will be kept after rotation. "
        "If set to zero, then no log files will be deleted. "
        "Defaults to 7.",
    )

    def apply_authentication(option, cli_values):
        if cli_values["admin_username"]:
            option.value["adminUsername"] = cli_values["admin_username"]

        if cli_values["password"]:
            option.value["password"] = cli_values["password"]

        if cli_values["password_file"]:
            option.value["passwordFile"] = cli_values["password_file"]

        if option.value.get("adminUsername"):
            password = option.value.get("password")
            passwordFile = option.value.get("passwordFile")
            if not password and not passwordFile:
                raise errors.InvalidConfiguration(
                    "Admin username specified without password."
                )
            if password and passwordFile:
                raise errors.InvalidConfiguration(
                    "Can't specify both password and password file."
                )

    default_authentication = {
        "adminUsername": None,
        "password": None,
        "passwordFile": None,
    }

    authentication = add_option(
        config_key="authentication",
        default=default_authentication,
        type=dict,
        apply_function=apply_authentication,
    )

    # -a is to specify the username for authentication.
    authentication.add_cli(
        "-a",
        "--admin-username",
        dest="admin_username",
        help="Used to specify the username of an admin user to "
        "authenticate with. To use authentication, the user "
        "must specify both an admin username and a keyFile.",
    )

    # -p is to specify the password used for authentication.
    authentication.add_cli(
        "-p",
        "--password",
        dest="password",
        help="Used to specify the password."
        " This is used by mongos to authenticate"
        " connections to the shards, and in the"
        " oplog threads. If authentication is not used, then"
        " this field can be left empty as the default ",
    )

    # -f is to specify the authentication key file. This file is used by mongos
    # to authenticate connections to the shards, and we'll use it in the oplog
    # threads.
    authentication.add_cli(
        "-f",
        "--password-file",
        dest="password_file",
        help="Used to store the password for authentication."
        " Use this option if you wish to specify a"
        " username and password but don't want to"
        " type in the password. The contents of this"
        " file should be the password for the admin user.",
    )

    def apply_fields(option, cli_values):
        if cli_values["fields"]:
            option.value = cli_values["fields"].split(",")

    fields = add_option(
        config_key="fields", default=[], type=list, apply_function=apply_fields
    )

    # -i to specify the list of fields to export
    fields.add_cli(
        "-i",
        "--fields",
        dest="fields",
        help="Use a comma separated list of fields to specify multiple fields. "
        "Will copy over the fields specified into a new document."
        "The '_id', 'ns' and '_ts' fields are always "
        "exported. Supports dot notation for document fields but cannot span "
        "arrays. Cannot use both 'fields' and 'exclude_fields'.",
    )

    def apply_exclude_fields(option, cli_values):
        if cli_values["exclude_fields"]:
            option.value = cli_values["exclude_fields"].split(",")

    exclude_fields = add_option(
        config_key="exclude_fields",
        default=[],
        type=list,
        apply_function=apply_exclude_fields,
    )

    # -i to specify the list of fields to exclude
    exclude_fields.add_cli(
        "-e",
        "--exclude_fields",
        dest="exclude_fields",
        help="Use a comma separated list of fields to specify multiple "
        "fields to exclude. Will delete the fields specified from the "
        "existing document. The '_id', 'ns' and '_ts' fields are always "
        "exported. Supports dot notation for document fields but cannot span "
        "arrays. Cannot use both 'fields' and 'exclude_fields'.",
    )

    def merge_namespaces_cli(option, cli_values):
        def validate_no_duplicates(lst, list_name):
            if len(lst) != len(set(lst)):
                raise errors.InvalidConfiguration(
                    "%s should not contain any duplicates." % (list_name,)
                )

        if cli_values["ns_set"]:
            option.value["include"] = cli_values["ns_set"].split(",")

        if cli_values["ex_ns_set"]:
            option.value["exclude"] = cli_values["ex_ns_set"].split(",")

        if cli_values["gridfs_set"]:
            option.value["gridfs"] = cli_values["gridfs_set"].split(",")

        if cli_values["dest_ns_set"]:
            ns_set = option.value["include"]
            dest_ns_set = cli_values["dest_ns_set"].split(",")
            if len(ns_set) != len(dest_ns_set):
                raise errors.InvalidConfiguration(
                    "Destination namespace set should be the"
                    " same length as the origin namespace set."
                )
            option.value["mapping"] = dict(zip(ns_set, dest_ns_set))

        # Check for duplicates
        if option.value["include"]:
            validate_no_duplicates(option.value["include"], "Include namespace set")
        if option.value["exclude"]:
            validate_no_duplicates(option.value["exclude"], "Exclude namespace set")
        if option.value["gridfs"]:
            validate_no_duplicates(option.value["gridfs"], "GridFS namespace set")
        if option.value["mapping"]:
            validate_no_duplicates(
                list(option.value["mapping"].values()), "Destination namespace set"
            )

    def apply_namespaces(option, cli_values):
        if (
            option.value["include"]
            or option.value["exclude"]
            or option.value["mapping"]
            or option.value["gridfs"]
        ):
            return apply_old_namespace_options(option, cli_values)
        else:
            return apply_new_namespace_options(option, cli_values)

    def apply_new_namespace_options(option, cli_values):
        """Apply the new format (since 2.5.0) namespaces options."""
        merge_namespaces_cli(option, cli_values)
        namespace_options = copy.deepcopy(option.value)
        ns_set = namespace_options.pop("include", None)
        ex_ns_set = namespace_options.pop("exclude", None)
        gridfs_set = namespace_options.pop("gridfs", None)
        dest_mapping = namespace_options.pop("mapping", None)
        option.value["namespace_options"] = namespace_options

        validate_namespace_options(
            namespace_set=ns_set,
            ex_namespace_set=ex_ns_set,
            dest_mapping=dest_mapping,
            namespace_options=namespace_options,
            gridfs_set=gridfs_set,
        )

    def apply_old_namespace_options(option, cli_values):
        """Apply the old format (before 2.5.0) namespaces options."""
        merge_namespaces_cli(option, cli_values)

        LOG.warning(
            "Deprecation warning: the current namespaces "
            "configuration format is outdated and support may be "
            "removed in a future release. Please update your "
            "config file to use the new format."
        )

        ns_set = option.value["include"]
        ex_ns_set = option.value["exclude"]
        gridfs_set = option.value["gridfs"]
        dest_mapping = option.value["mapping"]

        valid_names = set(["include", "exclude", "gridfs", "mapping"])
        valid_names |= set("__" + name for name in valid_names)
        valid_names.add("__comment__")
        for key in option.value:
            if key not in valid_names:
                raise errors.InvalidConfiguration(
                    "Invalid option %s in old style (pre 2.5.0) namespaces "
                    "configuration. The only valid option names are: %r"
                    % (key, list(valid_names))
                )

        validate_namespace_options(
            namespace_set=ns_set,
            ex_namespace_set=ex_ns_set,
            dest_mapping=dest_mapping,
            gridfs_set=gridfs_set,
        )

    default_namespaces = {"include": [], "exclude": [], "mapping": {}, "gridfs": []}

    namespaces = add_option(
        config_key="namespaces",
        default=default_namespaces,
        type=dict,
        apply_function=apply_namespaces,
    )

    # -n is to specify the namespaces we want to consider. The default
    # considers all the namespaces
    namespaces.add_cli(
        "-n",
        "--namespace-set",
        dest="ns_set",
        help="Used to specify the namespaces we want to "
        "consider. For example, if we wished to store all "
        "documents from the test.test and alpha.foo "
        "namespaces, we could use `-n test.test,alpha.foo`. "
        "You can also use, for example, `-n test.*` to store "
        "documents from all the collections of db test. "
        "The default is to consider all the namespaces, "
        "excluding the system and config databases, and "
        'also ignoring the "system.indexes" collection in '
        "any database.",
    )

    # -x is to specify the namespaces we dont want to consider. The default
    # is empty
    namespaces.add_cli(
        "-x",
        "--exclude-namespace-set",
        dest="ex_ns_set",
        help="Used to specify the namespaces we do not want to "
        "consider. For example, if we wished to ignore all "
        "documents from the test.test and alpha.foo "
        "namespaces, we could use `-x test.test,alpha.foo`. "
        "You can also use, for example, `-x test.*` to ignore "
        "documents from all the collections of db test. "
        "The default is not to exclude any namespace.",
    )

    # -g is the destination namespace
    namespaces.add_cli(
        "-g",
        "--dest-namespace-set",
        dest="dest_ns_set",
        help="Specify a destination namespace mapping. Each "
        "namespace provided in the --namespace-set option "
        "will be mapped respectively according to this "
        "comma-separated list. These lists must have "
        "equal length. "
        "It also supports mapping using wildcard, for example, "
        "mapping db.* to db.new_*, means that if we have two "
        "collections db.a and db.b, they will be renamed to "
        "db.new_a and db.new_b respectively. "
        "The default is to perform no renaming.",
    )

    # --gridfs-set is the set of GridFS namespaces to consider
    namespaces.add_cli(
        "--gridfs-set",
        dest="gridfs_set",
        help="Used to specify the GridFS namespaces we want to "
        "consider. For example, if your metadata is stored in "
        "test.fs.files and chunks are stored in test.fs.chunks, "
        "you can use `--gridfs-set test.fs`.",
    )

    def apply_doc_managers(option, cli_values):
        if not option.value:
            if not cli_values["doc_manager"] and not cli_values["target_url"]:
                return
            option.value = [{}]

        # Command line options should override the first DocManager config.
        cli_to_config = dict(
            doc_manager="docManager",
            target_url="targetURL",
            auto_commit_interval="autoCommitInterval",
            unique_key="uniqueKey",
        )
        first_dm = option.value[0]
        for cli_name, value in cli_values.items():
            if value is not None:
                first_dm[cli_to_config[cli_name]] = value

        # validate doc managers and fill in default values
        for dm in option.value:
            if not isinstance(dm, dict):
                raise errors.InvalidConfiguration(
                    "Elements of docManagers must be a dict."
                )
            if "docManager" not in dm and "docManagerClassPath" not in dm:
                raise errors.InvalidConfiguration(
                    "Every element of docManagers"
                    " must contain 'docManager' property."
                )
            if not dm.get("targetURL"):
                dm["targetURL"] = None
            if not dm.get("uniqueKey"):
                dm["uniqueKey"] = constants.DEFAULT_UNIQUE_KEY
            if dm.get("autoCommitInterval") is None:
                dm["autoCommitInterval"] = constants.DEFAULT_COMMIT_INTERVAL
            if not dm.get("args"):
                dm["args"] = {}
            if not dm.get("bulkSize"):
                dm["bulkSize"] = constants.DEFAULT_MAX_BULK

            aci = dm["autoCommitInterval"]
            if aci is not None and aci < 0:
                raise errors.InvalidConfiguration(
                    "autoCommitInterval must be non-negative."
                )

        def import_dm_by_name(name):
            full_name = "mongo_connector.doc_managers.%s.DocManager" % name
            return import_dm_by_path(full_name)

        def import_dm_by_path(path):
            # Avoid circular import on get_mininum_mongodb_version.
            from mongo_connector.doc_managers.doc_manager_base import DocManagerBase

            try:
                # importlib doesn't exist in 2.6, but __import__ is everywhere
                package, klass = path.rsplit(".", 1)
                module = __import__(package, fromlist=(package,))
                dm_impl = getattr(module, klass)
                if not issubclass(dm_impl, DocManagerBase):
                    raise TypeError("DocManager must inherit DocManagerBase.")
                return dm_impl
            except ImportError as exc:
                raise errors.InvalidConfiguration(
                    "Could not import %s. It could be that this doc manager ha"
                    "s been moved out of this project and is maintained elsewh"
                    "ere. Make sure that you have the doc manager installed al"
                    "ongside mongo-connector. Check the README for a list of a"
                    "vailable doc managers. ImportError:\n%s" % (package, exc)
                )
            except (AttributeError, TypeError):
                raise errors.InvalidConfiguration(
                    "No definition for DocManager found in %s." % package
                )

        # instantiate the doc manager objects
        dm_instances = []
        for dm in option.value:
            if "docManagerClassPath" in dm:
                DocManager = import_dm_by_path(dm["docManagerClassPath"])
            else:
                DocManager = import_dm_by_name(dm["docManager"])
            kwargs = {
                "unique_key": dm["uniqueKey"],
                "auto_commit_interval": dm["autoCommitInterval"],
                "chunk_size": dm["bulkSize"],
            }
            for k in dm["args"]:
                if k not in kwargs:
                    kwargs[k] = dm["args"][k]

            target_url = dm["targetURL"]
            if target_url:
                dm_instances.append(DocManager(target_url, **kwargs))
            else:
                dm_instances.append(DocManager(**kwargs))

        option.value = dm_instances

    doc_managers = add_option(
        config_key="docManagers",
        default=None,
        type=list,
        apply_function=apply_doc_managers,
    )

    # -d is to specify the doc manager file.
    doc_managers.add_cli(
        "-d",
        "--doc-manager",
        dest="doc_manager",
        help="Used to specify the path to each doc manager "
        "file that will be used. DocManagers should be "
        "specified in the same order as their respective "
        "target addresses in the --target-urls option. "
        "URLs are assigned to doc managers "
        "respectively. Additional doc managers are "
        "implied to have no target URL. Additional URLs "
        "are implied to have the same doc manager type as "
        "the last doc manager for which a URL was "
        "specified. By default, Mongo Connector will use "
        "'doc_manager_simulator.py'.  It is recommended "
        "that all doc manager files be kept in the "
        "doc_managers folder in mongo-connector. For "
        "more information about making your own doc "
        "manager, see 'Writing Your Own DocManager' "
        "section of the wiki",
    )

    # -d is to specify the doc manager file.
    doc_managers.add_cli(
        "-t",
        "--target-url",
        dest="target_url",
        help="Specify the URL to each target system being "
        "used. For example, if you were using Solr out of "
        "the box, you could use '-t "
        "http://localhost:8080/solr' with the "
        "SolrDocManager to establish a proper connection. "
        "URLs should be specified in the same order as "
        "their respective doc managers in the "
        "--doc-managers option.  URLs are assigned to doc "
        "managers respectively. Additional doc managers "
        "are implied to have no target URL. Additional "
        "URLs are implied to have the same doc manager "
        "type as the last doc manager for which a URL was "
        "specified. "
        "Don't use quotes around addresses. ",
    )

    # -u is to specify the mongoDB field that will serve as the unique key
    # for the target system,
    doc_managers.add_cli(
        "-u",
        "--unique-key",
        dest="unique_key",
        help="The name of the MongoDB field that will serve "
        "as the unique key for the target system. "
        "Note that this option does not apply "
        "when targeting another MongoDB cluster. "
        'Defaults to "_id".',
    )

    # --auto-commit-interval to specify auto commit time interval
    doc_managers.add_cli(
        "--auto-commit-interval",
        type="int",
        dest="auto_commit_interval",
        help="Seconds in-between calls for the Doc Manager"
        " to commit changes to the target system. A value of"
        " 0 means to commit after every write operation."
        " When left unset, Mongo Connector will not make"
        " explicit commits. Some systems have"
        " their own mechanism for adjusting a commit"
        " interval, which should be preferred to this"
        " option.",
    )

    continue_on_error = add_option(
        config_key="continueOnError", default=False, type=bool
    )

    def apply_ssl(option, cli_values):
        option.value = option.value or {}
        ssl_certfile = cli_values.get("ssl_certfile")
        if ssl_certfile is None:
            ssl_certfile = option.value.get("sslCertfile")
        ssl_keyfile = cli_values.get("ssl_keyfile")
        if ssl_keyfile is None:
            ssl_keyfile = option.value.get("sslKeyfile")
        ssl_ca_certs = cli_values.get("ssl_ca_certs")
        if ssl_ca_certs is None:
            ssl_ca_certs = option.value.get("sslCACerts")
        ssl_cert_reqs = cli_values.get("ssl_cert_reqs")
        if ssl_cert_reqs is None:
            ssl_cert_reqs = option.value.get("sslCertificatePolicy")

        if ssl_cert_reqs is not None:
            if ssl_cert_reqs not in _SSL_POLICY_MAP:
                raise errors.InvalidConfiguration(
                    "sslCertificatePolicy (--ssl-certificate-policy) must be "
                    'one of %s, got "%s"' % (_SSL_POLICY_MAP.keys(), ssl_cert_reqs)
                )
            if (
                pymongo.version_tuple < (3, 0)
                and ssl_cert_reqs != "ignored"
                and not ssl_ca_certs
            ):
                raise errors.InvalidConfiguration(
                    '--ssl-certificate-policy is not "ignored" and '
                    "--ssl-ca-certs was not be provided. Either upgrade "
                    "PyMongo to >= 3.0 to load system provided CA "
                    "certificates or specify a CA file with --ssl-ca-certs."
                )

        option.value["sslCertfile"] = ssl_certfile
        option.value["sslCACerts"] = ssl_ca_certs
        option.value["sslKeyfile"] = ssl_keyfile
        option.value["sslCertificatePolicy"] = _SSL_POLICY_MAP.get(ssl_cert_reqs)

    ssl = add_option(config_key="ssl", default={}, type=dict, apply_function=apply_ssl)
    ssl.add_cli(
        "--ssl-certfile",
        dest="ssl_certfile",
        help=("Path to a certificate identifying the local connection " "to MongoDB."),
    )
    ssl.add_cli(
        "--ssl-keyfile",
        dest="ssl_keyfile",
        help=(
            "Path to the private key for --ssl-certfile. "
            "Not necessary if already included in --ssl-certfile."
        ),
    )
    ssl.add_cli(
        "--ssl-certificate-policy",
        dest="ssl_cert_reqs",
        choices=("required", "optional", "ignored"),
        help=(
            "Policy for validating SSL certificates provided from the other "
            "end of the connection. There are three possible values: "
            "required = Require and validate the remote certificate. "
            'optional = The same as "required", unless the server was '
            "configured to use anonymous ciphers. "
            "ignored = Remote SSL certificates are ignored completely."
        ),
    )
    ssl.add_cli(
        "--ssl-ca-certs",
        dest="ssl_ca_certs",
        help=(
            "Path to a concatenated set of certificate authority "
            "certificates to validate the other side of the connection. "
        ),
    )

    # --continue-on-error to continue to upsert documents during a collection
    # dump, even if the documents cannot be inserted for some reason
    continue_on_error.add_cli(
        "--continue-on-error",
        action="store_true",
        dest="continue_on_error",
        help="By default, if any document fails to upsert"
        " during a collection dump, the entire operation fails."
        " When this flag is enabled, normally fatal errors"
        " will be caught and logged, allowing the collection"
        " dump to continue.\n"
        "Note: Applying oplog operations to an incomplete"
        " set of documents due to errors may cause undefined"
        " behavior. Use this flag to dump only.",
    )

    config_file = add_option()
    config_file.add_cli(
        "-c",
        "--config-file",
        dest="config_file",
        help="Specify a JSON file to load configurations from. You can find"
        " an example config file at mongo-connector/config.json",
    )

    tz_aware = add_option(config_key="timezoneAware", default=False, type=bool)
    tz_aware.add_cli(
        "--tz-aware",
        dest="tz_aware",
        action="store_true",
        help="Make all dates and times timezone-aware.",
    )

    return result


def setup_logging(conf):
    root_logger = logging.getLogger()
    formatter = logging.Formatter(conf["logging.format"])

    log_levels = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
    loglevel = log_levels[conf["verbosity"]]
    root_logger.setLevel(loglevel)

    if conf["logging.type"] == "file":
        log_out = logging.handlers.TimedRotatingFileHandler(
            conf["logging.filename"],
            when=conf["logging.rotationWhen"],
            interval=conf["logging.rotationInterval"],
            backupCount=conf["logging.rotationBackups"],
        )
        print("Logging to %s." % conf["logging.filename"])
    elif conf["logging.type"] == "syslog":
        syslog_info = conf["logging.host"]
        if ":" in syslog_info:
            log_host, log_port = syslog_info.split(":")
            syslog_info = (log_host, int(log_port))
        log_out = logging.handlers.SysLogHandler(
            address=syslog_info, facility=conf["logging.facility"]
        )
        print("Logging to system log at %s" % conf["logging.host"])
    elif conf["logging.type"] == "stream":
        log_out = logging.StreamHandler()
    else:
        print(
            "Logging type must be one of 'stream', 'syslog', or 'file', not "
            "'%s'." % conf["logging.type"]
        )
        sys.exit(1)

    log_out.setLevel(loglevel)
    log_out.setFormatter(formatter)
    root_logger.addHandler(log_out)
    return root_logger


def log_startup_info():
    """Log info about the current environment."""
    LOG.always("Starting mongo-connector version: %s", __version__)
    if "dev" in __version__:
        LOG.warning(
            "This is a development version (%s) of mongo-connector", __version__
        )
    LOG.always("Python version: %s", sys.version)
    LOG.always("Platform: %s", platform.platform())
    if hasattr(pymongo, "__version__"):
        pymongo_version = pymongo.__version__
    else:
        pymongo_version = pymongo.version
    LOG.always("pymongo version: %s", pymongo_version)
    if not pymongo.has_c():
        LOG.warning(
            "pymongo version %s was installed without the C extensions. "
            '"InvalidBSON: Date value out of range" errors may occur if '
            "there are documents with BSON Datetimes that represent times "
            "outside of Python's datetime limit.",
            pymongo.__version__,
        )


@log_fatal_exceptions
def main():
    """ Starts the mongo connector (assuming CLI)
    """
    # Setup an initial logging handler that buffers log messages before
    # applying the final logging configuration.
    initial_handler = logging.handlers.MemoryHandler(100)
    root_logger = logging.getLogger()
    root_logger.addHandler(initial_handler)

    # Parse configuration and setup logging.
    conf = config.Config(get_config_options())
    conf.parse_args()
    setup_logging(conf)

    # Flush the buffered log messages to the final logging handler.
    initial_handler.setTarget(root_logger.handlers[-1])
    initial_handler.flush()
    root_logger.removeHandler(initial_handler)

    log_startup_info()

    connector = Connector.from_config(conf)

    # Catch SIGTERM and SIGINT to cleanup the connector gracefully
    def signame_handler(signal_name):
        def sig_handler(signum, frame):
            # Save the signal so it can be printed later
            connector.signal = (signal_name, signum)
            connector.can_run = False

        return sig_handler

    signal.signal(signal.SIGTERM, signame_handler("SIGTERM"))
    signal.signal(signal.SIGINT, signame_handler("SIGINT"))

    connector.start()

    while True:
        if not connector.is_alive():
            break
        time.sleep(3)


if __name__ == "__main__":
    main()
